import math

import torch
from nnsvs.base import BaseModel
from nnsvs.util import make_pad_mask
from torch import nn
from torch.nn import Parameter
from torch.nn import functional as F


def softmax(x, dim):
    return F.softmax(x, dim=dim, dtype=torch.float32)


class PositionalEncoding(torch.nn.Module):
    """Positional encoding.
    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_len (int): Maximum input length.
        reverse (bool): Whether to reverse the input position.
    """

    def __init__(self, d_model, dropout_rate, max_len=5000, reverse=False):
        """Construct an PositionalEncoding object."""
        super(PositionalEncoding, self).__init__()
        self.d_model = d_model
        self.reverse = reverse
        self.xscale = math.sqrt(self.d_model)
        self.dropout = torch.nn.Dropout(p=dropout_rate)
        self.pe = None
        self.extend_pe(torch.tensor(0.0).expand(1, max_len))

    def extend_pe(self, x):
        """Reset the positional encodings."""
        if self.pe is not None:
            if self.pe.size(1) >= x.size(1):
                if self.pe.dtype != x.dtype or self.pe.device != x.device:
                    self.pe = self.pe.to(dtype=x.dtype, device=x.device)
                return
        pe = torch.zeros(x.size(1), self.d_model)
        if self.reverse:
            position = torch.arange(
                x.size(1) - 1, -1, -1.0, dtype=torch.float32
            ).unsqueeze(1)
        else:
            position = torch.arange(0, x.size(1), dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, dtype=torch.float32)
            * -(math.log(10000.0) / self.d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.pe = pe.to(device=x.device, dtype=x.dtype)

    def forward(self, x: torch.Tensor):
        """Add positional encoding.
        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).
        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).
        """
        self.extend_pe(x)
        x = x * self.xscale + self.pe[:, : x.size(1)]
        return self.dropout(x)


class RelPositionalEncoding(PositionalEncoding):
    """Relative positional encoding module.
    See : Appendix B in https://arxiv.org/abs/1901.02860
    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_len (int): Maximum input length.
    """

    def __init__(self, d_model, dropout_rate, max_len=5000):
        """Initialize class."""
        super().__init__(d_model, dropout_rate, max_len, reverse=True)

    def forward(self, x):
        """Compute positional encoding.
        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).
        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).
            torch.Tensor: Positional embedding tensor (1, time, `*`).
        """
        self.extend_pe(x)
        x = x * self.xscale
        pos_emb = self.pe[:, : x.size(1)]
        return self.dropout(x) + self.dropout(pos_emb)


class MultiheadAttention(nn.Module):
    def __init__(
        self,
        embed_dim,
        num_heads,
        kdim=None,
        vdim=None,
        dropout=0.0,
        bias=True,
        add_bias_kv=False,
        add_zero_attn=False,
        self_attention=False,
        encoder_decoder_attention=False,
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self.qkv_same_dim = self.kdim == embed_dim and self.vdim == embed_dim

        self.num_heads = num_heads
        self.dropout = dropout
        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"
        self.scaling = self.head_dim ** -0.5

        self.self_attention = self_attention
        self.encoder_decoder_attention = encoder_decoder_attention

        assert not self.self_attention or self.qkv_same_dim, (
            "Self-attention requires query, key and " "value to be of the same size"
        )

        if self.qkv_same_dim:
            self.in_proj_weight = Parameter(torch.Tensor(3 * embed_dim, embed_dim))
        else:
            self.k_proj_weight = Parameter(torch.Tensor(embed_dim, self.kdim))
            self.v_proj_weight = Parameter(torch.Tensor(embed_dim, self.vdim))
            self.q_proj_weight = Parameter(torch.Tensor(embed_dim, embed_dim))

        if bias:
            self.in_proj_bias = Parameter(torch.Tensor(3 * embed_dim))
        else:
            self.register_parameter("in_proj_bias", None)

        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)

        if add_bias_kv:
            self.bias_k = Parameter(torch.Tensor(1, 1, embed_dim))
            self.bias_v = Parameter(torch.Tensor(1, 1, embed_dim))
        else:
            self.bias_k = self.bias_v = None

        self.add_zero_attn = add_zero_attn

        self.reset_parameters()

        self.enable_torch_version = False
        if hasattr(F, "multi_head_attention_forward"):
            self.enable_torch_version = True
        else:
            self.enable_torch_version = False
        self.last_attn_probs = None

    def reset_parameters(self):
        if self.qkv_same_dim:
            nn.init.xavier_uniform_(self.in_proj_weight)
        else:
            nn.init.xavier_uniform_(self.k_proj_weight)
            nn.init.xavier_uniform_(self.v_proj_weight)
            nn.init.xavier_uniform_(self.q_proj_weight)

        nn.init.xavier_uniform_(self.out_proj.weight)
        if self.in_proj_bias is not None:
            nn.init.constant_(self.in_proj_bias, 0.0)
            nn.init.constant_(self.out_proj.bias, 0.0)
        if self.bias_k is not None:
            nn.init.xavier_normal_(self.bias_k)
        if self.bias_v is not None:
            nn.init.xavier_normal_(self.bias_v)

    def forward(
        self,
        query,
        key,
        value,
        key_padding_mask=None,
        incremental_state=None,
        need_weights=True,
        static_kv=False,
        attn_mask=None,
        before_softmax=False,
        need_head_weights=False,
        enc_dec_attn_constraint_mask=None,
        reset_attn_weight=None,
    ):
        """Input shape: Time x Batch x Channel

        Args:
            key_padding_mask (ByteTensor, optional): mask to exclude
                keys that are pads, of shape `(batch, src_len)`, where
                padding elements are indicated by 1s.
            need_weights (bool, optional): return the attention weights,
                averaged over heads (default: False).
            attn_mask (ByteTensor, optional): typically used to
                implement causal attention, where the mask prevents the
                attention from looking forward in time (default: None).
            before_softmax (bool, optional): return the raw attention
                weights and values before the attention softmax.
            need_head_weights (bool, optional): return the attention
                weights for each head. Implies *need_weights*. Default:
                return the average attention weights over all heads.
        """
        if need_head_weights:
            need_weights = True

        tgt_len, bsz, embed_dim = query.size()
        assert embed_dim == self.embed_dim
        assert list(query.size()) == [tgt_len, bsz, embed_dim]

        if (
            self.enable_torch_version
            and incremental_state is None
            and not static_kv
            and reset_attn_weight is None
        ):
            if self.qkv_same_dim:
                return F.multi_head_attention_forward(
                    query,
                    key,
                    value,
                    self.embed_dim,
                    self.num_heads,
                    self.in_proj_weight,
                    self.in_proj_bias,
                    self.bias_k,
                    self.bias_v,
                    self.add_zero_attn,
                    self.dropout,
                    self.out_proj.weight,
                    self.out_proj.bias,
                    self.training,
                    key_padding_mask,
                    need_weights,
                    attn_mask,
                )
            else:
                return F.multi_head_attention_forward(
                    query,
                    key,
                    value,
                    self.embed_dim,
                    self.num_heads,
                    torch.empty([0]),
                    self.in_proj_bias,
                    self.bias_k,
                    self.bias_v,
                    self.add_zero_attn,
                    self.dropout,
                    self.out_proj.weight,
                    self.out_proj.bias,
                    self.training,
                    key_padding_mask,
                    need_weights,
                    attn_mask,
                    use_separate_proj_weight=True,
                    q_proj_weight=self.q_proj_weight,
                    k_proj_weight=self.k_proj_weight,
                    v_proj_weight=self.v_proj_weight,
                )

        if incremental_state is not None:
            raise NotImplementedError()
        else:
            saved_state = None

        if self.self_attention:
            # self-attention
            q, k, v = self.in_proj_qkv(query)
        elif self.encoder_decoder_attention:
            # encoder-decoder attention
            q = self.in_proj_q(query)
            if key is None:
                assert value is None
                k = v = None
            else:
                k = self.in_proj_k(key)
                v = self.in_proj_v(key)

        else:
            q = self.in_proj_q(query)
            k = self.in_proj_k(key)
            v = self.in_proj_v(value)
        q *= self.scaling

        if self.bias_k is not None:
            assert self.bias_v is not None
            k = torch.cat([k, self.bias_k.repeat(1, bsz, 1)])
            v = torch.cat([v, self.bias_v.repeat(1, bsz, 1)])
            if attn_mask is not None:
                attn_mask = torch.cat(
                    [attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1
                )
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [
                        key_padding_mask,
                        key_padding_mask.new_zeros(key_padding_mask.size(0), 1),
                    ],
                    dim=1,
                )

        q = (
            q.contiguous()
            .view(tgt_len, bsz * self.num_heads, self.head_dim)
            .transpose(0, 1)
        )
        if k is not None:
            k = (
                k.contiguous()
                .view(-1, bsz * self.num_heads, self.head_dim)
                .transpose(0, 1)
            )
        if v is not None:
            v = (
                v.contiguous()
                .view(-1, bsz * self.num_heads, self.head_dim)
                .transpose(0, 1)
            )

        if saved_state is not None:
            raise NotImplementedError()

        src_len = k.size(1)

        # This is part of a workaround to get around fork/join parallelism
        # not supporting Optional types.
        if key_padding_mask is not None and key_padding_mask.shape == torch.Size([]):
            key_padding_mask = None

        if key_padding_mask is not None:
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        if self.add_zero_attn:
            src_len += 1
            k = torch.cat([k, k.new_zeros((k.size(0), 1) + k.size()[2:])], dim=1)
            v = torch.cat([v, v.new_zeros((v.size(0), 1) + v.size()[2:])], dim=1)
            if attn_mask is not None:
                attn_mask = torch.cat(
                    [attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1
                )
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [
                        key_padding_mask,
                        torch.zeros(key_padding_mask.size(0), 1).type_as(
                            key_padding_mask
                        ),
                    ],
                    dim=1,
                )

        attn_weights = torch.bmm(q, k.transpose(1, 2))
        attn_weights = self.apply_sparse_mask(attn_weights, tgt_len, src_len, bsz)

        assert list(attn_weights.size()) == [bsz * self.num_heads, tgt_len, src_len]

        if attn_mask is not None:
            if len(attn_mask.shape) == 2:
                attn_mask = attn_mask.unsqueeze(0)
            elif len(attn_mask.shape) == 3:
                attn_mask = (
                    attn_mask[:, None]
                    .repeat([1, self.num_heads, 1, 1])
                    .reshape(bsz * self.num_heads, tgt_len, src_len)
                )
            attn_weights = attn_weights + attn_mask

        if enc_dec_attn_constraint_mask is not None:  # bs x head x L_kv
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            attn_weights = attn_weights.masked_fill(
                enc_dec_attn_constraint_mask.unsqueeze(2).bool(),
                -1e9,
            )
            attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        if key_padding_mask is not None:
            # don't attend to padding symbols
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            attn_weights = attn_weights.masked_fill(
                key_padding_mask.unsqueeze(1).unsqueeze(2),
                -1e9,
            )
            attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        attn_logits = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)

        if before_softmax:
            return attn_weights, v

        attn_weights_float = softmax(attn_weights, dim=-1)
        attn_weights = attn_weights_float.type_as(attn_weights)
        attn_probs = F.dropout(
            attn_weights_float.type_as(attn_weights),
            p=self.dropout,
            training=self.training,
        )

        if reset_attn_weight is not None:
            if reset_attn_weight:
                self.last_attn_probs = attn_probs.detach()
            else:
                assert self.last_attn_probs is not None
                attn_probs = self.last_attn_probs
        attn = torch.bmm(attn_probs, v)
        assert list(attn.size()) == [bsz * self.num_heads, tgt_len, self.head_dim]
        attn = attn.transpose(0, 1).contiguous().view(tgt_len, bsz, embed_dim)
        attn = self.out_proj(attn)

        if need_weights:
            attn_weights = attn_weights_float.view(
                bsz, self.num_heads, tgt_len, src_len
            ).transpose(1, 0)
            if not need_head_weights:
                # average attention weights over heads
                attn_weights = attn_weights.mean(dim=0)
        else:
            attn_weights = None

        return attn, (attn_weights, attn_logits)

    def in_proj_qkv(self, query):
        return self._in_proj(query).chunk(3, dim=-1)

    def in_proj_q(self, query):
        if self.qkv_same_dim:
            return self._in_proj(query, end=self.embed_dim)
        else:
            bias = self.in_proj_bias
            if bias is not None:
                bias = bias[: self.embed_dim]
            return F.linear(query, self.q_proj_weight, bias)

    def in_proj_k(self, key):
        if self.qkv_same_dim:
            return self._in_proj(key, start=self.embed_dim, end=2 * self.embed_dim)
        else:
            weight = self.k_proj_weight
            bias = self.in_proj_bias
            if bias is not None:
                bias = bias[self.embed_dim : 2 * self.embed_dim]
            return F.linear(key, weight, bias)

    def in_proj_v(self, value):
        if self.qkv_same_dim:
            return self._in_proj(value, start=2 * self.embed_dim)
        else:
            weight = self.v_proj_weight
            bias = self.in_proj_bias
            if bias is not None:
                bias = bias[2 * self.embed_dim :]
            return F.linear(value, weight, bias)

    def _in_proj(self, input, start=0, end=None):
        weight = self.in_proj_weight
        bias = self.in_proj_bias
        weight = weight[start:end, :]
        if bias is not None:
            bias = bias[start:end]
        return F.linear(input, weight, bias)

    def apply_sparse_mask(self, attn_weights, tgt_len, src_len, bsz):
        return attn_weights


def LayerNorm(normalized_shape, eps=1e-5, elementwise_affine=True):
    return torch.nn.LayerNorm(normalized_shape, eps, elementwise_affine)


class BatchNorm1dTBC(nn.Module):
    def __init__(self, c):
        super(BatchNorm1dTBC, self).__init__()
        self.bn = nn.BatchNorm1d(c)

    def forward(self, x):
        """

        :param x: [T, B, C]
        :return: [T, B, C]
        """
        x = x.permute(1, 2, 0)  # [B, C, T]
        x = self.bn(x)  # [B, C, T]
        x = x.permute(2, 0, 1)  # [T, B, C]
        return x


def Linear(in_features, out_features, bias=True):
    m = nn.Linear(in_features, out_features, bias)
    nn.init.xavier_uniform_(m.weight)
    if bias:
        nn.init.constant_(m.bias, 0.0)
    return m


class Swish(torch.autograd.Function):
    @staticmethod
    def forward(ctx, i):
        result = i * torch.sigmoid(i)
        ctx.save_for_backward(i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        i = ctx.saved_variables[0]
        sigmoid_i = torch.sigmoid(i)
        return grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i)))


class CustomSwish(nn.Module):
    def forward(self, input_tensor):
        return Swish.apply(input_tensor)


class TransformerFFNLayer(nn.Module):
    def __init__(
        self,
        hidden_size,
        filter_size,
        padding="SAME",
        kernel_size=1,
        dropout=0.0,
        act="gelu",
    ):
        super().__init__()
        self.kernel_size = kernel_size
        self.dropout = dropout
        self.act = act
        if padding == "SAME":
            self.ffn_1 = nn.Conv1d(
                hidden_size, filter_size, kernel_size, padding=kernel_size // 2
            )
        elif padding == "LEFT":
            self.ffn_1 = nn.Sequential(
                nn.ConstantPad1d((kernel_size - 1, 0), 0.0),
                nn.Conv1d(hidden_size, filter_size, kernel_size),
            )
        self.ffn_2 = Linear(filter_size, hidden_size)
        if self.act == "swish":
            self.swish_fn = CustomSwish()

    def forward(self, x, incremental_state=None):
        # x: T x B x C
        if incremental_state is not None:
            assert incremental_state is None, "Nar-generation does not allow this."

        x = self.ffn_1(x.permute(1, 2, 0)).permute(2, 0, 1)
        x = x * self.kernel_size ** -0.5

        if incremental_state is not None:
            x = x[-1:]
        if self.act == "gelu":
            x = F.gelu(x)
        if self.act == "relu":
            x = F.relu(x)
        if self.act == "swish":
            x = self.swish_fn(x)
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.ffn_2(x)
        return x


class EncSALayer(nn.Module):
    def __init__(
        self,
        c,
        num_heads,
        dropout,
        attention_dropout=0.1,
        relu_dropout=0.1,
        kernel_size=9,
        padding="SAME",
        norm="ln",
        act="gelu",
    ):
        super().__init__()
        self.c = c
        self.dropout = dropout
        self.num_heads = num_heads
        if num_heads > 0:
            if norm == "ln":
                self.layer_norm1 = LayerNorm(c)
            elif norm == "bn":
                self.layer_norm1 = BatchNorm1dTBC(c)
            self.self_attn = MultiheadAttention(
                self.c,
                num_heads,
                self_attention=True,
                dropout=attention_dropout,
                bias=False,
            )
        if norm == "ln":
            self.layer_norm2 = LayerNorm(c)
        elif norm == "bn":
            self.layer_norm2 = BatchNorm1dTBC(c)
        self.ffn = TransformerFFNLayer(
            c,
            4 * c,
            kernel_size=kernel_size,
            dropout=relu_dropout,
            padding=padding,
            act=act,
        )

    def forward(self, x, encoder_padding_mask=None, **kwargs):
        layer_norm_training = kwargs.get("layer_norm_training", None)
        if layer_norm_training is not None:
            self.layer_norm1.training = layer_norm_training
            self.layer_norm2.training = layer_norm_training
        if self.num_heads > 0:
            residual = x
            x = self.layer_norm1(x)
            x, _, = self.self_attn(
                query=x, key=x, value=x, key_padding_mask=encoder_padding_mask
            )
            x = F.dropout(x, self.dropout, training=self.training)
            x = residual + x
            x = x * (1 - encoder_padding_mask.float()).transpose(0, 1)[..., None]

        residual = x
        x = self.layer_norm2(x)
        x = self.ffn(x)
        x = F.dropout(x, self.dropout, training=self.training)
        x = residual + x
        x = x * (1 - encoder_padding_mask.float()).transpose(0, 1)[..., None]
        return x


class TransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        hidden_size,
        dropout,
        kernel_size=None,
        num_heads=2,
        norm="ln",
        padding="SAME",
        act="gelu",
    ):
        super().__init__()
        self.hidden_size = hidden_size
        self.dropout = dropout
        self.num_heads = num_heads
        self.op = EncSALayer(
            hidden_size,
            num_heads,
            dropout=dropout,
            attention_dropout=0.0,
            relu_dropout=dropout,
            kernel_size=kernel_size,
            padding=padding,
            norm=norm,
            act=act,
        )

    def forward(self, x, **kwargs):
        return self.op(x, **kwargs)


class FFTBlocks(nn.Module):
    def __init__(
        self,
        hidden_size,
        num_layers,
        ffn_kernel_size=9,
        dropout=0.1,
        num_heads=2,
        use_pos_embed=True,
        use_last_norm=True,
        norm="ln",
        use_pos_embed_alpha=True,
    ):
        super().__init__()
        self.num_layers = num_layers
        embed_dim = self.hidden_size = hidden_size
        self.dropout = dropout
        self.use_pos_embed = use_pos_embed
        self.use_last_norm = use_last_norm
        if use_pos_embed:
            self.padding_idx = 0
            self.pos_embed_alpha = (
                nn.Parameter(torch.Tensor([1])) if use_pos_embed_alpha else 1
            )
            self.embed_positions = RelPositionalEncoding(hidden_size, dropout_rate=0.0)

        self.layers = nn.ModuleList([])
        self.layers.extend(
            [
                TransformerEncoderLayer(
                    self.hidden_size,
                    self.dropout,
                    kernel_size=ffn_kernel_size,
                    num_heads=num_heads,
                )
                for _ in range(self.num_layers)
            ]
        )
        if self.use_last_norm:
            if norm == "ln":
                self.layer_norm = nn.LayerNorm(embed_dim)
            elif norm == "bn":
                self.layer_norm = BatchNorm1dTBC(embed_dim)
        else:
            self.layer_norm = None

    def forward(self, x, lengths, padding_mask=None):
        """
        :param x: [B, T, C]
        :param padding_mask: [B, T]
        :return: [B, T, C] or [L, B, T, C]
        """
        if padding_mask is None:
            padding_mask = make_pad_mask(lengths).to(x.device)

        # padding_mask = x.abs().sum(-1).eq(0).data if padding_mask is None else padding_mask
        nonpadding_mask_TB = (
            1 - padding_mask.transpose(0, 1).float()[:, :, None]
        )  # [T, B, 1]
        if self.use_pos_embed:
            # positions = self.pos_embed_alpha * self.embed_positions(x[..., 0])
            positions = self.pos_embed_alpha * self.embed_positions(x)
            x = x + positions
            x = F.dropout(x, p=self.dropout, training=self.training)
        # B x T x C -> T x B x C
        x = x.transpose(0, 1) * nonpadding_mask_TB
        hiddens = []
        for layer in self.layers:
            x = layer(x, encoder_padding_mask=padding_mask) * nonpadding_mask_TB
            hiddens.append(x)
        if self.use_last_norm:
            x = self.layer_norm(x) * nonpadding_mask_TB
        x = x.transpose(0, 1)  # [B, T, C]
        return x


class FFTBlocksEncoder(BaseModel):
    def __init__(
        self,
        in_dim,
        out_dim,
        hidden_dim,
        num_layers=2,
        ffn_kernel_size=9,
        dropout=0.1,
        num_heads=2,
        use_pos_embed=True,
        use_last_norm=True,
        norm="ln",
        use_pos_embed_alpha=True,
        reduction_factor=1,
        downsample_by_conv=True,
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim
        self.reduction_factor = reduction_factor

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            self.fc = nn.Linear(embed_dim, hidden_dim)
        else:
            self.emb = None
            self.fc_in = None
            self.fc = nn.Linear(in_dim, hidden_dim)

        if reduction_factor > 1 and downsample_by_conv:
            self.conv_downsample = nn.Conv1d(
                in_dim,
                in_dim,
                kernel_size=reduction_factor,
                stride=reduction_factor,
                groups=in_dim,
            )
        else:
            self.conv_downsample = None

        self.encoder = FFTBlocks(
            hidden_size=hidden_dim,
            num_layers=num_layers,
            ffn_kernel_size=ffn_kernel_size,
            dropout=dropout,
            num_heads=num_heads,
            use_pos_embed=use_pos_embed,
            use_last_norm=use_last_norm,
            norm=norm,
            use_pos_embed_alpha=use_pos_embed_alpha,
        )
        self.fc_out = nn.Linear(hidden_dim, out_dim * reduction_factor)

    def forward(self, x, lengths, y=None):
        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        # Adjust lengths based on the reduction factor
        if self.reduction_factor > 1:
            lengths = (lengths / self.reduction_factor).long()
            if self.conv_downsample is not None:
                x = self.conv_downsample(x.transpose(1, 2)).transpose(1, 2)
            else:
                x = x[:, self.reduction_factor - 1 :: self.reduction_factor]

        x = self.fc(x)
        x = self.encoder(x, lengths)

        x = self.fc_out(x).view(x.shape[0], -1, self.out_dim)

        return x



from pathlib import Path

import pytest
from nnmnkwii.datasets import FileSourceDataset
from nnsvs.data import MelF0AcousticSource, WORLDAcousticSource


@pytest.mark.skipif(
    not (Path(__file__).parent / "data" / "utt_list.txt").exists(),
    reason="test data not found",
)
@pytest.mark.parametrize("f0_extractor", ["dio", "harvest", "parselmouth"])
def test_world(f0_extractor):
    cwd = Path(__file__).parent
    utt_list = cwd / "data" / "utt_list.txt"
    wav_root = cwd / "data" / "wav"
    label_root = cwd / "data" / "label_phone_align"
    hed_path = cwd / "data" / "test.hed"
    data_source = WORLDAcousticSource(
        utt_list=utt_list,
        wav_root=wav_root,
        label_root=label_root,
        question_path=hed_path,
        dynamic_features_flags=[False, False, False, False],
        relative_f0=False,
        sample_rate=24000,
        f0_extractor=f0_extractor,
        num_windows=1,
    )

    dataset = FileSourceDataset(data_source)
    for data in dataset:
        y, wave, y_pf = data
        print(y.shape, wave.shape, y_pf.shape)
        break


@pytest.mark.skipif(
    not (Path(__file__).parent / "data" / "utt_list.txt").exists(),
    reason="test data not found",
)
@pytest.mark.parametrize("f0_extractor", ["dio", "harvest", "parselmouth"])
def test_melf0(f0_extractor):
    cwd = Path(__file__).parent
    utt_list = cwd / "data" / "utt_list.txt"
    wav_root = cwd / "data" / "wav"
    label_root = cwd / "data" / "label_phone_align"
    hed_path = cwd / "data" / "test.hed"
    data_source = MelF0AcousticSource(
        utt_list=utt_list,
        wav_root=wav_root,
        label_root=label_root,
        question_path=hed_path,
        sample_rate=24000,
        fft_size=512,
        hop_size=120,
        win_length=480,
        f0_extractor=f0_extractor,
    )

    dataset = FileSourceDataset(data_source)
    for data in dataset:
        y, wave, y_pf = data
        print(y.shape, wave.shape, y_pf.shape)
        break



"""This module provides functionality for pitch analysis.

References:

Nakano et al, "An Automatic Singing Skill Evaluation Method
for Unknown Melodies Using Pitch Interval Accuracy and Vibrato Features"
Proc. Interspeech 2006.

山田 et al, "HMM に基づく歌声合成のためのビブラートモデル化"
IPSJ SIG Tech. Report 2009.

Note that vibrato extraction method in this module is exerimental.
Because details of the vibrato extraction method are not described
in the above papers and not trivial to implement (in my opinion),
my implementation may not work well compared to the original author's one.
Also note that there are a lot of tunable parameters (threshold,
window size, min/max extent, cut-off frequency, etc.).
If you want to get maximum performance, you might want to tune these
parameters with your dataset.
I tested this code with kiritan_singing and nit-song070 database.
"""
import librosa
import numpy as np
import torch
from nnsvs.dsp import lowpass_filter
from scipy.signal import argrelmax, argrelmin

_c4_hz = 440 * 2 ** (3 / 12 - 1)
_c4_cent = 4800


def hz_to_cent_based_c4(hz):
    """Convert Hz to cent based on C4

    Args:
        hz (np.ndarray): array of Hz

    Returns:
        np.ndarray: array of cent
    """
    out = hz.copy()
    nonzero_indices = np.where(hz > 0)[0]
    out[nonzero_indices] = (
        1200 * np.log(hz[nonzero_indices] / _c4_hz) / np.log(2) + _c4_cent
    )
    return out


def cent_to_hz_based_c4(cent):
    """Convert cent to Hz based on C4

    Args:
        cent (np.ndarray): array of cent

    Returns:
        np.ndarray: array of Hz
    """
    out = cent.copy()
    nonzero_indices = np.where(cent > 0)[0]
    out[nonzero_indices] = (
        np.exp((cent[nonzero_indices] - _c4_cent) * np.log(2) / 1200) * _c4_hz
    )
    return out


def nonzero_segments(f0):
    """Find nonzero segments

    Args:
        f0 (np.ndarray): array of f0

    Returns:
        list: list of (start, end)
    """
    vuv = f0 > 0
    started = False
    s, e = 0, 0
    segments = []
    for idx in range(len(f0)):
        if vuv[idx] > 0 and not started:
            started = True
            s = idx
        elif started and (vuv[idx] <= 0):
            e = idx
            started = False
            segments.append((s, e))
        else:
            pass

    if started and vuv[-1] > 0:
        segments.append((s, len(vuv) - 1))

    return segments


def note_segments(lf0_score_denorm):
    """Compute note segments (start and end indices) from log-F0

    Note that unvoiced frames must be set to 0 in advance.

    Args:
        lf0_score_denorm (Tensor): (B, T)

    Returns:
        list: list of note (start, end) indices
    """
    segments = []
    for s, e in nonzero_segments(lf0_score_denorm):
        out = torch.sign(torch.abs(torch.diff(lf0_score_denorm[s : e + 1])))
        transitions = torch.where(out > 0)[0]
        note_start, note_end = s, -1
        for pos in transitions:
            note_end = int(s + pos)
            segments.append((note_start, note_end))
            note_start = note_end + 1

        # Handle last note
        while (
            note_start < len(lf0_score_denorm) - 1 and lf0_score_denorm[note_start] <= 0
        ):
            note_start += 1
        note_end = note_start + 1
        while note_end < len(lf0_score_denorm) - 1 and lf0_score_denorm[note_end] > 0:
            note_end += 1

        if note_end != note_start + 1:
            segments.append((note_start, note_end))

    return segments


def compute_f0_correction_ratio(
    f0,
    f0_score,
    edges_to_be_excluded=50,
    out_of_tune_threshold=200,
    correction_threshold=100,
):
    """Compute f0 correction ratio

    Args:
        f0 (np.ndarray): array of f0
        f0_score (np.ndarray): array of f0 score

    Returns:
        float: correction ratio to multiplied to F0 (i.e. f0 * ratio)
    """
    segments = note_segments(torch.from_numpy(f0_score))

    center_f0s = []
    center_score_f0s = []
    # edges_to_be_excluded = 50  # 0.25 sec for excluding overshoot/preparation
    for s, e in segments:
        L = e - s
        if L > edges_to_be_excluded * 2:
            center_f0s.append(f0[s + edges_to_be_excluded : e - edges_to_be_excluded])
            center_score_f0s.append(
                f0_score[s + edges_to_be_excluded : e - edges_to_be_excluded]
            )
    center_f0s = np.concatenate(center_f0s)
    center_score_f0s = np.concatenate(center_score_f0s)

    # Compute pitch ratio to be multiplied
    nonzero_indices = (center_f0s > 0) & (center_score_f0s > 0)
    ratio = center_score_f0s[nonzero_indices] / center_f0s[nonzero_indices]

    # Exclude too out-of-tune frames (over 2 semitone)
    up_threshold = np.exp(out_of_tune_threshold * np.log(2) / 1200)
    low_threshold = np.exp(-out_of_tune_threshold * np.log(2) / 1200)

    ratio = ratio[(ratio < up_threshold) & (ratio > low_threshold)]

    global_offset = ratio.mean()

    # Avoid corrections over semi-tone
    # If more than semi-tone pitch correction is needed, it is better to correct
    # data by hand or fix musicxml or UST instead.
    up_threshold = np.exp(correction_threshold * np.log(2) / 1200)
    low_threshold = np.exp(-correction_threshold * np.log(2) / 1200)

    if global_offset > up_threshold or global_offset < low_threshold:
        print(
            f"""warn: more than 1 semitone pitch correction is needed.
global_offset: {global_offset} cent.
It is likely that manual pitch corrections are preferable."""
        )
    global_offset = np.clip(global_offset, low_threshold, up_threshold)

    return global_offset


def extract_vibrato_parameters_impl(pitch_seg, sr):
    """Extract vibrato parameters for a single pitch segment

    Nakano et al, "An Automatic Singing Skill Evaluation Method
    for Unknown Melodies Using Pitch Interval Accuracy and Vibrato Features"
    Proc. Interspeech 2006.

    山田 et al, "HMM に基づく歌声合成のためのビブラートモデル化"
    IPSJ SIG Tech. Report 2009.

    Args:
        pitch_seg (np.ndarray): array of pitch
        sr (int): sampling rate

    Returns:
        tuple: (R, E, m_a, m_f)
    """
    peak_high_pos = argrelmax(pitch_seg)[0]
    peak_low_pos = argrelmin(pitch_seg)[0]

    m_a = np.zeros(len(pitch_seg))
    m_f = np.zeros(len(pitch_seg))

    if len(peak_high_pos) != len(peak_low_pos) + 1:
        print("Warning! Probably a bug...T.T")
        print(peak_high_pos, peak_low_pos)
        return None, None, None, None

    peak_high_pos_diff = np.diff(peak_high_pos)
    peak_low_pos_diff = np.diff(peak_low_pos)

    R = np.zeros(len(peak_high_pos_diff) + len(peak_low_pos_diff))
    R[0::2] = peak_high_pos_diff
    R[1::2] = peak_low_pos_diff

    m_f_ind = np.zeros(len(R), dtype=int)
    m_f_ind[0::2] = peak_high_pos[:-1]
    m_f_ind[1::2] = peak_low_pos[:-1]
    m_f[m_f_ind] = (1 / R) * sr

    peak_high_pitch = pitch_seg[peak_high_pos]
    peak_low_pitch = pitch_seg[peak_low_pos]

    E = np.zeros(len(R))
    E[0::2] = (peak_high_pitch[1:] + peak_high_pitch[:-1]) / 2 - peak_low_pitch
    E[1::2] = peak_high_pitch[1:-1] - (peak_low_pitch[1:] + peak_low_pitch[:-1]) / 2

    m_a_ind = np.zeros(len(R), dtype=int)
    m_a_ind[0::2] = peak_low_pos
    m_a_ind[1::2] = peak_high_pos[1:-1]
    m_a[m_a_ind] = 0.5 * E

    rate = 1 / R.mean() * sr
    extent = 0.5 * E.mean()
    print(f"Rate: {rate}, Extent: {extent}")

    return R, E, m_a, m_f


def compute_extent(pitch_seg):
    """Compute extent of a pitch segment

    Args:
        pitch_seg (np.ndarray): array of pitch

    Returns:
        np.ndarray: array of extent
    """
    peak_high_pos = argrelmax(pitch_seg)[0]
    peak_low_pos = argrelmin(pitch_seg)[0]

    if len(peak_high_pos) == 1 or len(peak_low_pos) == 1:
        return np.array([-1])

    if len(peak_high_pos) < len(peak_low_pos):
        peak_low_pos = peak_low_pos[:-2]
    elif len(peak_high_pos) == len(peak_low_pos):
        peak_low_pos = peak_low_pos[:-1]

    peak_high_pitch = pitch_seg[peak_high_pos]
    peak_low_pitch = pitch_seg[peak_low_pos]

    peak_high_pos_diff = np.diff(peak_high_pos)
    peak_low_pos_diff = np.diff(peak_low_pos)

    # TODO: would probably be a bug...
    if len(peak_high_pitch) != len(peak_low_pitch) + 1:
        return np.array([-1])

    E = np.zeros(len(peak_high_pos_diff) + len(peak_low_pos_diff))
    E[0::2] = (peak_high_pitch[1:] + peak_high_pitch[:-1]) / 2 - peak_low_pitch
    E[1::2] = peak_high_pitch[1:-1] - (peak_low_pitch[1:] + peak_low_pitch[:-1]) / 2

    return E


def extract_smoothed_f0(f0, sr, cutoff=8):
    """Extract smoothed f0 by low-pass filtering

    Note that the low-pass filter is only applied to voiced segments.

    Args:
        f0 (np.ndarray): array of f0
        sr (int): sampling rate
        cutoff (float): cutoff frequency

    Returns:
        np.ndarray: array of smoothed f0
    """
    segments = nonzero_segments(f0)

    f0_smooth = f0.copy()
    for s, e in segments:
        f0_smooth[s:e] = lowpass_filter(f0[s:e], sr, cutoff=cutoff)

    return f0_smooth


def extract_smoothed_continuous_f0(f0, sr, cutoff=20):
    """Extract smoothed continuous f0 by low-pass filtering

    Note that the input must be continuous F0 or log-F0.

    Args:
        f0 (np.ndarray): array of continuous f0
        sr (int): sampling rate
        cutoff (float): initial cutoff frequency

    Returns:
        np.ndarray: array of smoothed continuous f0
    """
    is_2d = len(f0.shape) == 2
    f0 = f0.reshape(-1) if is_2d else f0

    # Ref: https://bit.ly/3SOePFw
    f0_smooth = lowpass_filter(f0, sr, cutoff=cutoff)

    # Fallback case: shound't happen I believe
    # NOTE: hard-coded for now
    next_cutoff = 50
    while (f0_smooth < 0).any():
        f0_smooth = lowpass_filter(f0, sr, cutoff=next_cutoff)
        next_cutoff *= 2

    f0_smooth = f0_smooth.reshape(len(f0), 1) if is_2d else f0_smooth
    return f0_smooth


def extract_vibrato_likelihood(
    f0_smooth, sr, win_length=32, n_fft=128, min_freq=3, max_freq=8
):
    """Extract vibrato likelihood

    Args:
        f0_smooth (np.ndarray): array of smoothed f0
        sr (int): sampling rate
        win_length (int): window length
        n_fft (int): FFT size
        min_freq (float): minimum frequency of the vibrato
        max_freq (float): maximum frequency of the vibrato

    Returns:
        np.ndarray: array of vibrato likelihood
    """
    # STFT on 1st order diffference of F0
    X = np.abs(
        librosa.stft(
            np.diff(f0_smooth),
            hop_length=1,
            win_length=win_length,
            n_fft=n_fft,
            window="hann",
        )
    )
    X_norm = X / (X.sum(0) + 1e-7)

    freq_per_bin = sr / n_fft
    min_freq_bin = int(min_freq / freq_per_bin)
    max_freq_bin = int(max_freq / freq_per_bin)

    # Compute vibrato likelhiood
    St = np.abs(np.diff(X_norm, axis=0)).sum(0)
    Ft = X_norm[min_freq_bin:max_freq_bin, :].sum(0)
    vibrato_likelihood = St * Ft

    return vibrato_likelihood


def interp_vibrato(m_f):
    """Interpolate a sequence of vibrato parameter by linear interpolation

    Args:
        m_f (np.ndarray): array of vibrato parameter

    Returns:
        np.ndarray: array of vibrato parameter
    """
    nonzero_indices = np.where(m_f > 0)[0]
    nonzero_indices = [0] + list(nonzero_indices) + [len(m_f) - 1]
    out = np.interp(np.arange(len(m_f)), nonzero_indices, m_f[nonzero_indices])
    return out


def extract_vibrato_parameters(
    pitch,
    vibrato_likelihood,
    sr=100,
    threshold=0.12,
    min_cross_count=5,
    min_extent=30,
    max_extent=150,
    interp_params=True,
    smooth_params=False,
    smooth_width=15,
    clip_extent=True,
):
    """Extract vibrato parameters

    Args:
        pitch (np.ndarray): array of pitch (smoothed f0)
        vibrato_likelihood (np.ndarray): array of vibrato likelihood
        sr (int): sampling rate
        threshold (float): threshold of vibrato likelihood
        min_cross_count (int): minimum number of cross points
        min_extent (int): minimum extent of vibrato (cent)
        max_extent (int): maximum extent of vibrato (cent)
        interp_params (bool): whether to interpolate vibrato parameters
        smooth_params (bool): whether to smooth vibrato parameters
        smooth_width (int): width of smoothing window
        clip_extent (bool): whether to clip extent

    Returns:
        tuple: tuple of vibrato parameters
    """
    T = len(vibrato_likelihood)

    vibrato_flags = np.zeros(T, dtype=int)
    m_a = np.zeros(T)
    m_f = np.zeros(T)

    peak_high_pos = argrelmax(pitch)[0]
    peak_low_pos = argrelmin(pitch)[0]

    # iterate over every peak position
    peak_high_idx = 0
    while peak_high_idx < len(peak_high_pos):
        peak_frame_idx = peak_high_pos[peak_high_idx]

        found = False
        if vibrato_likelihood[peak_frame_idx] > threshold:
            # Initial positions for vibrato section
            start_index = peak_frame_idx
            peaks = peak_low_pos[peak_low_pos > peak_frame_idx]
            if len(peaks) > 0:
                end_index = peaks[0]
            else:
                peak_high_idx += 1
                continue
            next_start_peak_high_idx = -1

            # Find a peak position that is close to the next non-speech segment
            # assuming that there's a non-speech segment right after vibrato
            # NOTE: we may want to remove this constraint
            peak_high_pos_rest = peak_high_pos[peak_high_pos > peak_frame_idx]
            for frame_idx in range(end_index, T):
                if pitch[frame_idx] <= 0:
                    peaks = peak_high_pos_rest[peak_high_pos_rest < frame_idx]
                    if len(peaks) > 0:
                        end_index = peaks[-1]
                        next_start_peak_high_idx = (
                            len(peak_high_pos[peak_high_pos < end_index]) + 1
                        )
                    break

            # Set the search width (backward)
            search_width_backward = 0
            for frame_idx in range(start_index, 0, -1):
                if pitch[frame_idx] <= 0:
                    peaks_backward = peak_high_pos[
                        (peak_high_pos < peak_frame_idx) & (peak_high_pos > frame_idx)
                    ]
                    if len(peaks_backward) > 0:
                        backward = peaks_backward[0]
                        search_width_backward = len(
                            peak_high_pos[
                                (peak_high_pos > backward)
                                & (peak_high_pos <= peak_frame_idx)
                            ]
                        )
                    break

            # Find a peak position that satisfies the following vibrato constraints
            # 1) more than 5 times crossing
            # 2) 30 ~ 150 cent oscillation
            estimate_start_index = start_index
            rate = 0
            for peak_idx in range(
                max(peak_high_idx - search_width_backward, 0), peak_high_idx
            ):
                if peak_high_pos[peak_idx] >= T:
                    break
                f0_seg = pitch[peak_high_pos[peak_idx] : end_index]

                # Check if the segment satisfies vibrato constraints
                m = f0_seg.mean()
                cross_count = len(np.where(np.diff(np.sign(f0_seg - m)))[0])

                # Find the start_index so that the vibrato section has more than 5 crossing
                E = compute_extent(f0_seg)
                extent = 0.5 * E.mean()
                having_large_deviation = ((0.5 * E) > max_extent * 2).any()
                if (
                    cross_count >= min_cross_count
                    and cross_count >= rate
                    and extent >= min_extent
                    and extent <= max_extent
                    and not having_large_deviation
                    and (E > 0).all()
                ):
                    rate = cross_count
                    estimate_start_index = peak_high_pos[peak_idx]

            start_index = estimate_start_index

            if rate >= min_cross_count:
                R, E, m_a_seg, m_f_seg = extract_vibrato_parameters_impl(
                    pitch[start_index - 1 : end_index + 2], sr
                )
                if m_a_seg is None:
                    found = False
                    break
                found = True
                vibrato_flags[start_index:end_index] = 1

                if interp_params:
                    m_a_seg = interp_vibrato(m_a_seg)
                    m_f_seg = np.clip(interp_vibrato(m_f_seg), 3, 8)
                if smooth_params:
                    m_a_seg = np.convolve(
                        m_a_seg, np.ones(smooth_width) / smooth_width, mode="same"
                    )
                    m_f_seg = np.convolve(
                        m_f_seg, np.ones(smooth_width) / smooth_width, mode="same"
                    )

                if clip_extent:
                    m_a_seg = np.clip(m_a_seg, min_extent, max_extent)
                m_a[start_index:end_index] = m_a_seg[1:-2]
                m_f[start_index:end_index] = m_f_seg[1:-2]

                assert next_start_peak_high_idx > peak_high_idx
                peak_high_idx = next_start_peak_high_idx

        if not found:
            peak_high_idx += 1

    return vibrato_flags, m_a, m_f


def gen_sine_vibrato(f0, sr, m_a, m_f, scale=1.0):
    """Generate F0 with sine-based vibrato

    Args:
        f0 (ndarray): fundamental frequency
        sr (int): sampling rate
        m_a (ndarray): amplitude of vibrato
        m_f (ndarray): frequency of vibrato
        scale (float): scale factor

    Returns:
        ndarray: F0 with sine-based vibrato
    """
    f0_gen = f0.copy()

    voiced_end_indices = np.asarray([e for _, e in nonzero_segments(f0)])

    for s, e in nonzero_segments(m_a):
        # limit vibrato rate to [3, 8] Hz
        m_f_seg = np.clip(m_f[s:e], 3, 8)
        # limit vibrato extent to [30, 150] cent
        m_a_seg = np.clip(m_a[s:e], 30, 150)

        cent = scale * m_a_seg * np.sin(2 * np.pi / sr * m_f_seg * np.arange(0, e - s))
        new_f0 = f0[s:e] * np.exp(cent * np.log(2) / 1200)
        f0_gen[s:e] = new_f0

        # NOTE: this is a hack to avoid discontinuity at the end of vibrato
        voiced_ends_next_to_vibrato = voiced_end_indices[voiced_end_indices > e]
        if len(voiced_ends_next_to_vibrato) > 0:
            voiced_end = voiced_ends_next_to_vibrato[0]
            f0_gen[s:voiced_end] = lowpass_filter(f0_gen[s:voiced_end], sr, cutoff=12)

    return f0_gen



import os
from os.path import join

import hydra
import joblib
import numpy as np
import torch
from hydra.utils import to_absolute_path
from nnmnkwii.io import hts
from nnsvs.gen import (
    postprocess_acoustic,
    postprocess_waveform,
    predict_acoustic,
    predict_timing,
    predict_waveform,
)
from nnsvs.logger import getLogger
from nnsvs.util import extract_static_scaler, init_seed, load_utt_list, load_vocoder
from omegaconf import DictConfig, OmegaConf
from scipy.io import wavfile
from tqdm.auto import tqdm


@hydra.main(config_path="conf/synthesis", config_name="config")
def my_app(config: DictConfig) -> None:
    global logger
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    if not torch.cuda.is_available():
        device = torch.device("cpu")
    else:
        device = torch.device(config.device)

    # timelag
    timelag_config = OmegaConf.load(to_absolute_path(config.timelag.model_yaml))
    timelag_model = hydra.utils.instantiate(timelag_config.netG).to(device)
    checkpoint = torch.load(
        to_absolute_path(config.timelag.checkpoint),
        map_location=lambda storage, loc: storage,
    )
    timelag_model.load_state_dict(checkpoint["state_dict"])
    timelag_in_scaler = joblib.load(to_absolute_path(config.timelag.in_scaler_path))
    timelag_out_scaler = joblib.load(to_absolute_path(config.timelag.out_scaler_path))
    timelag_model.eval()

    # duration
    duration_config = OmegaConf.load(to_absolute_path(config.duration.model_yaml))
    duration_model = hydra.utils.instantiate(duration_config.netG).to(device)
    checkpoint = torch.load(
        to_absolute_path(config.duration.checkpoint),
        map_location=lambda storage, loc: storage,
    )
    duration_model.load_state_dict(checkpoint["state_dict"])
    duration_in_scaler = joblib.load(to_absolute_path(config.duration.in_scaler_path))
    duration_out_scaler = joblib.load(to_absolute_path(config.duration.out_scaler_path))
    duration_model.eval()

    # acoustic model
    acoustic_config = OmegaConf.load(to_absolute_path(config.acoustic.model_yaml))
    acoustic_model = hydra.utils.instantiate(acoustic_config.netG).to(device)
    checkpoint = torch.load(
        to_absolute_path(config.acoustic.checkpoint),
        map_location=lambda storage, loc: storage,
    )
    acoustic_model.load_state_dict(checkpoint["state_dict"])
    acoustic_in_scaler = joblib.load(to_absolute_path(config.acoustic.in_scaler_path))
    acoustic_out_scaler = joblib.load(to_absolute_path(config.acoustic.out_scaler_path))
    acoustic_model.eval()

    # NOTE: this is used for GV post-filtering
    acoustic_out_static_scaler = extract_static_scaler(
        acoustic_out_scaler, acoustic_config
    )

    # Vocoder
    if config.vocoder.checkpoint is not None and len(config.vocoder.checkpoint) > 0:
        vocoder, vocoder_in_scaler, vocoder_config = load_vocoder(
            to_absolute_path(config.vocoder.checkpoint),
            device,
            acoustic_config,
        )
    else:
        vocoder, vocoder_in_scaler, vocoder_config = None, None, None
        if config.synthesis.vocoder_type != "world":
            logger.warning("Vocoder checkpoint is not specified")
            logger.info(f"Use world instead of {config.synthesis.vocoder_type}.")
        config.synthesis.vocoder_type = "world"

    # Run synthesis for each utt.
    binary_dict, numeric_dict = hts.load_question_set(
        to_absolute_path(config.synthesis.qst)
    )

    in_dir = to_absolute_path(config.in_dir)
    out_dir = to_absolute_path(config.out_dir)
    os.makedirs(out_dir, exist_ok=True)
    utt_ids = load_utt_list(to_absolute_path(config.utt_list))
    logger.info("Processes %s utterances...", len(utt_ids))
    for utt_id in tqdm(utt_ids):
        labels = hts.load(join(in_dir, f"{utt_id}.lab"))
        hts_frame_shift = int(config.synthesis.frame_period * 1e4)
        labels.frame_shift = hts_frame_shift
        init_seed(1234)

        if config.synthesis.ground_truth_duration:
            duration_modified_labels = labels
        else:
            duration_modified_labels = predict_timing(
                device=device,
                labels=labels,
                binary_dict=binary_dict,
                numeric_dict=numeric_dict,
                timelag_model=timelag_model,
                timelag_config=timelag_config,
                timelag_in_scaler=timelag_in_scaler,
                timelag_out_scaler=timelag_out_scaler,
                duration_model=duration_model,
                duration_config=duration_config,
                duration_in_scaler=duration_in_scaler,
                duration_out_scaler=duration_out_scaler,
                log_f0_conditioning=config.synthesis.log_f0_conditioning,
                allowed_range=config.timelag.allowed_range,
                allowed_range_rest=config.timelag.allowed_range_rest,
                force_clip_input_features=config.timelag.force_clip_input_features,
                frame_period=config.synthesis.frame_period,
            )

        # Predict acoustic features
        acoustic_features = predict_acoustic(
            device=device,
            labels=duration_modified_labels,
            acoustic_model=acoustic_model,
            acoustic_config=acoustic_config,
            acoustic_in_scaler=acoustic_in_scaler,
            acoustic_out_scaler=acoustic_out_scaler,
            binary_dict=binary_dict,
            numeric_dict=numeric_dict,
            subphone_features=config.synthesis.subphone_features,
            log_f0_conditioning=config.synthesis.log_f0_conditioning,
            force_clip_input_features=config.acoustic.force_clip_input_features,
            f0_shift_in_cent=config.synthesis.pre_f0_shift_in_cent,
        )

        # NOTE: the output of this function is tuple of features
        # e.g., (mgc, lf0, vuv, bap)
        multistream_features = postprocess_acoustic(
            device=device,
            acoustic_features=acoustic_features,
            duration_modified_labels=duration_modified_labels,
            binary_dict=binary_dict,
            numeric_dict=numeric_dict,
            acoustic_config=acoustic_config,
            acoustic_out_static_scaler=acoustic_out_static_scaler,
            postfilter_model=None,  # NOTE: learned post-filter is not supported
            postfilter_config=None,
            postfilter_out_scaler=None,
            sample_rate=config.synthesis.sample_rate,
            frame_period=config.synthesis.frame_period,
            relative_f0=config.synthesis.relative_f0,
            feature_type=config.synthesis.feature_type,
            post_filter_type=config.synthesis.post_filter_type,
            trajectory_smoothing=config.synthesis.trajectory_smoothing,
            trajectory_smoothing_cutoff=config.synthesis.trajectory_smoothing_cutoff,
            trajectory_smoothing_cutoff_f0=config.synthesis.trajectory_smoothing_cutoff_f0,
            vuv_threshold=config.synthesis.vuv_threshold,
            f0_shift_in_cent=config.synthesis.post_f0_shift_in_cent,
            vibrato_scale=1.0,
            force_fix_vuv=config.synthesis.force_fix_vuv,
        )

        # Generate waveform by vocoder
        wav = predict_waveform(
            device=device,
            multistream_features=multistream_features,
            vocoder=vocoder,
            vocoder_config=vocoder_config,
            vocoder_in_scaler=vocoder_in_scaler,
            sample_rate=config.synthesis.sample_rate,
            frame_period=config.synthesis.frame_period,
            use_world_codec=config.synthesis.use_world_codec,
            feature_type=config.synthesis.feature_type,
            vocoder_type=config.synthesis.vocoder_type,
            vuv_threshold=config.synthesis.vuv_threshold,
        )

        wav = postprocess_waveform(
            wav=wav,
            sample_rate=config.synthesis.sample_rate,
            dtype=np.int16,
            peak_norm=False,
            loudness_norm=False,
        )

        out_wav_path = join(out_dir, f"{utt_id}.wav")
        wavfile.write(
            out_wav_path, rate=config.synthesis.sample_rate, data=wav.astype(np.int16)
        )


def entry():
    my_app()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



# coding: utf-8

import os
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from glob import glob
from itertools import zip_longest
from os.path import basename, exists, join, splitext

import hydra
import joblib
import numpy as np
from hydra.utils import to_absolute_path
from nnsvs.logger import getLogger
from omegaconf import DictConfig, OmegaConf
from sklearn.preprocessing import StandardScaler
from tqdm import tqdm

logger = None


def get_paths_by_glob(in_dir, filt):
    return glob(join(in_dir, filt))


def _process_utterance(
    out_dir, audio_path, feat_path, scaler, inverse, clipping_threshold
):
    # [Optional] copy audio with the same name if exists
    if audio_path is not None and exists(audio_path):
        name = splitext(basename(audio_path))[0]
        np.save(join(out_dir, name), np.load(audio_path), allow_pickle=False)

    # [Required] apply normalization for features
    assert exists(feat_path)
    x = np.load(feat_path)
    if inverse:
        y = scaler.inverse_transform(x)
    else:
        y = scaler.transform(x)

    # NOTE: workaround for outliers
    # may need to adjust the clipping value
    if isinstance(scaler, StandardScaler):
        y = np.clip(y, -clipping_threshold, clipping_threshold)

    assert x.dtype == y.dtype
    name = splitext(basename(feat_path))[0]
    np.save(join(out_dir, name), y, allow_pickle=False)


def apply_normalization_dir2dir(
    in_dir, out_dir, scaler, inverse, num_workers, clipping_threshold
):
    # NOTE: at this point, audio_paths can be empty
    audio_paths = get_paths_by_glob(in_dir, "*-wave.npy")
    feature_paths = get_paths_by_glob(in_dir, "*-feats.npy")
    executor = ProcessPoolExecutor(max_workers=num_workers)
    futures = []
    for audio_path, feature_path in zip_longest(audio_paths, feature_paths):
        futures.append(
            executor.submit(
                partial(
                    _process_utterance,
                    out_dir,
                    audio_path,
                    feature_path,
                    scaler,
                    inverse,
                    clipping_threshold,
                )
            )
        )
    for future in tqdm(futures):
        future.result()


@hydra.main(config_path="conf/preprocess_normalize", config_name="config")
def my_app(config: DictConfig) -> None:
    global logger
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    in_dir = to_absolute_path(config.in_dir)
    out_dir = to_absolute_path(config.out_dir)
    scaler_path = to_absolute_path(config.scaler_path)
    scaler = joblib.load(scaler_path)
    inverse = config.inverse
    num_workers = config.num_workers
    clipping_threshold = config.clipping_threshold

    os.makedirs(out_dir, exist_ok=True)
    apply_normalization_dir2dir(
        in_dir, out_dir, scaler, inverse, num_workers, clipping_threshold
    )


def entry():
    my_app()


if __name__ == "__main__":
    my_app()



from collections import deque
from functools import partial

import numpy as np
import torch
from nnsvs.base import BaseModel, PredictionType
from tqdm import tqdm


def extract(a, t, x_shape):
    b, *_ = t.shape
    out = a.gather(-1, t)
    return out.reshape(b, *((1,) * (len(x_shape) - 1)))


def noise_like(shape, noise_fn, device, repeat=False):
    if repeat:
        resid = [1] * (len(shape) - 1)
        shape_one = (1, *shape[1:])

        return noise_fn(*shape_one, device=device).repeat(shape[0], *resid)

    else:
        return noise_fn(*shape, device=device)


def linear_beta_schedule(timesteps, min_beta=1e-4, max_beta=0.06):
    """
    linear schedule
    """
    betas = np.linspace(min_beta, max_beta, timesteps)
    return betas


def cosine_beta_schedule(timesteps, s=0.008):
    """
    cosine schedule
    as proposed in https://openreview.net/forum?id=-NEXDKk8gZ
    """
    steps = timesteps + 1
    x = np.linspace(0, steps, steps)
    alphas_cumprod = np.cos(((x / steps) + s) / (1 + s) * np.pi * 0.5) ** 2
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return np.clip(betas, a_min=0, a_max=0.999)


beta_schedule = {
    "cosine": cosine_beta_schedule,
    "linear": linear_beta_schedule,
}


class GaussianDiffusion(BaseModel):
    def __init__(
        self,
        in_dim,
        out_dim,
        denoise_fn,
        encoder=None,
        K_step=100,
        betas=None,
        schedule_type="linear",
        scheduler_params=None,
        norm_scale=10,
        pndm_speedup=None,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.denoise_fn = denoise_fn
        self.K_step = K_step
        self.pndm_speedup = pndm_speedup
        self.encoder = encoder
        self.norm_scale = norm_scale
        if scheduler_params is None:
            if schedule_type == "linear":
                scheduler_params = {"max_beta": 0.06}
            elif schedule_type == "cosine":
                scheduler_params = {"s": 0.008}

        if encoder is not None:
            assert encoder.in_dim == in_dim, "encoder input dim must match in_dim"
        assert out_dim == denoise_fn.in_dim, "denoise_fn input dim must match out_dim"

        if pndm_speedup:
            raise NotImplementedError("pndm_speedup is not implemented yet")

        if betas is not None:
            betas = (
                betas.detach().cpu().numpy()
                if isinstance(betas, torch.Tensor)
                else betas
            )
        else:
            betas = beta_schedule[schedule_type](K_step, **scheduler_params)

        alphas = 1.0 - betas
        alphas_cumprod = np.cumprod(alphas, axis=0)
        alphas_cumprod_prev = np.append(1.0, alphas_cumprod[:-1])

        self.noise_list = deque(maxlen=4)

        to_torch = partial(torch.tensor, dtype=torch.float32)
        self.register_buffer("betas", to_torch(betas))
        self.register_buffer("alphas_cumprod", to_torch(alphas_cumprod))
        self.register_buffer("alphas_cumprod_prev", to_torch(alphas_cumprod_prev))

        # calculations for diffusion q(x_t | x_{t-1}) and others
        self.register_buffer("sqrt_alphas_cumprod", to_torch(np.sqrt(alphas_cumprod)))
        self.register_buffer(
            "sqrt_one_minus_alphas_cumprod", to_torch(np.sqrt(1.0 - alphas_cumprod))
        )
        self.register_buffer(
            "log_one_minus_alphas_cumprod", to_torch(np.log(1.0 - alphas_cumprod))
        )
        self.register_buffer(
            "sqrt_recip_alphas_cumprod", to_torch(np.sqrt(1.0 / alphas_cumprod))
        )
        self.register_buffer(
            "sqrt_recipm1_alphas_cumprod", to_torch(np.sqrt(1.0 / alphas_cumprod - 1))
        )

        # calculations for posterior q(x_{t-1} | x_t, x_0)
        posterior_variance = (
            betas * (1.0 - alphas_cumprod_prev) / (1.0 - alphas_cumprod)
        )
        # above: equal to 1. / (1. / (1. - alpha_cumprod_tm1) + alpha_t / beta_t)
        self.register_buffer("posterior_variance", to_torch(posterior_variance))
        # below: log calculation clipped because the posterior variance is 0
        # at the beginning of the diffusion chain
        self.register_buffer(
            "posterior_log_variance_clipped",
            to_torch(np.log(np.maximum(posterior_variance, 1e-20))),
        )
        self.register_buffer(
            "posterior_mean_coef1",
            to_torch(betas * np.sqrt(alphas_cumprod_prev) / (1.0 - alphas_cumprod)),
        )
        self.register_buffer(
            "posterior_mean_coef2",
            to_torch(
                (1.0 - alphas_cumprod_prev) * np.sqrt(alphas) / (1.0 - alphas_cumprod)
            ),
        )

    # NOTE: since the original impl. assume the data is distributed in [-1, 1]
    # let us (roughly) convert N(0,1) noramlized to data to [-1, 1]
    def _norm(self, x, a_max=10):
        return x / a_max

    def _denorm(self, x, a_max=10):
        return x * a_max

    def prediction_type(self):
        return PredictionType.DIFFUSION

    def q_mean_variance(self, x_start, t):
        mean = extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
        variance = extract(1.0 - self.alphas_cumprod, t, x_start.shape)
        log_variance = extract(self.log_one_minus_alphas_cumprod, t, x_start.shape)
        return mean, variance, log_variance

    def predict_start_from_noise(self, x_t, t, noise):
        return (
            extract(self.sqrt_recip_alphas_cumprod, t, x_t.shape) * x_t
            - extract(self.sqrt_recipm1_alphas_cumprod, t, x_t.shape) * noise
        )

    def q_posterior(self, x_start, x_t, t):
        posterior_mean = (
            extract(self.posterior_mean_coef1, t, x_t.shape) * x_start
            + extract(self.posterior_mean_coef2, t, x_t.shape) * x_t
        )
        posterior_variance = extract(self.posterior_variance, t, x_t.shape)
        posterior_log_variance_clipped = extract(
            self.posterior_log_variance_clipped, t, x_t.shape
        )
        return posterior_mean, posterior_variance, posterior_log_variance_clipped

    def p_mean_variance(self, x, t, cond, clip_denoised: bool):
        noise_pred = self.denoise_fn(x, t, cond=cond)
        x_recon = self.predict_start_from_noise(x, t=t, noise=noise_pred)

        if clip_denoised:
            x_recon.clamp_(-1.0, 1.0)

        model_mean, posterior_variance, posterior_log_variance = self.q_posterior(
            x_start=x_recon, x_t=x, t=t
        )
        return model_mean, posterior_variance, posterior_log_variance

    @torch.no_grad()
    def p_sample(
        self, x, t, cond, noise_fn=torch.randn, clip_denoised=True, repeat_noise=False
    ):
        b, *_, device = *x.shape, x.device
        model_mean, _, model_log_variance = self.p_mean_variance(
            x=x, t=t, cond=cond, clip_denoised=clip_denoised
        )
        noise = noise_like(x.shape, noise_fn, device, repeat_noise)
        # no noise when t == 0
        nonzero_mask = (1 - (t == 0).float()).reshape(b, *((1,) * (len(x.shape) - 1)))
        return model_mean + nonzero_mask * (0.5 * model_log_variance).exp() * noise

    @torch.no_grad()
    def p_sample_plms(self, x, t, interval, cond):
        """
        Use the PLMS method from Pseudo Numerical Methods for Diffusion Models on Manifolds
        https://arxiv.org/abs/2202.09778.
        """

        def get_x_pred(x, noise_t, t):
            a_t = extract(self.alphas_cumprod, t, x.shape)
            a_prev = extract(
                self.alphas_cumprod,
                torch.max(t - interval, torch.zeros_like(t)),
                x.shape,
            )
            a_t_sq, a_prev_sq = a_t.sqrt(), a_prev.sqrt()

            x_delta = (a_prev - a_t) * (
                (1 / (a_t_sq * (a_t_sq + a_prev_sq))) * x
                - 1
                / (a_t_sq * (((1 - a_prev) * a_t).sqrt() + ((1 - a_t) * a_prev).sqrt()))
                * noise_t
            )
            x_pred = x + x_delta

            return x_pred

        noise_list = self.noise_list
        noise_pred = self.denoise_fn(x, t, cond=cond)

        if len(noise_list) == 0:
            x_pred = get_x_pred(x, noise_pred, t)
            # noise_pred_prev = self.denoise_fn(x_pred, max(t - interval, 0), cond=cond)
            noise_pred_prev = self.denoise_fn(
                x_pred, torch.max(t - interval, torch.zeros_like(t)), cond=cond
            )
            noise_pred_prime = (noise_pred + noise_pred_prev) / 2
        elif len(noise_list) == 1:
            noise_pred_prime = (3 * noise_pred - noise_list[-1]) / 2
        elif len(noise_list) == 2:
            noise_pred_prime = (
                23 * noise_pred - 16 * noise_list[-1] + 5 * noise_list[-2]
            ) / 12
        elif len(noise_list) >= 3:
            noise_pred_prime = (
                55 * noise_pred
                - 59 * noise_list[-1]
                + 37 * noise_list[-2]
                - 9 * noise_list[-3]
            ) / 24

        x_prev = get_x_pred(x, noise_pred_prime, t)
        noise_list.append(noise_pred)

        return x_prev

    def q_sample(self, x_start, t, noise=None):
        if noise is None:
            noise = torch.randn_like(x_start)
        return (
            extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
            + extract(self.sqrt_one_minus_alphas_cumprod, t, x_start.shape) * noise
        )

    def forward(self, cond, lengths=None, y=None):
        """Forward step

        Args:
            cond (torch.Tensor): conditioning features of shaep (B, T, encoder_hidden_dim)
            lengths (torch.Tensor): lengths of each sequence in the batch
            y (torch.Tensor): ground truth of shape (B, T, C)

        Returns:
            tuple of tensors (B, T, in_dim), (B, T, in_dim)
        """
        B = cond.shape[0]
        device = cond.device

        if self.encoder is not None:
            cond = self.encoder(cond, lengths)

        # (B, M, T)
        cond = cond.transpose(1, 2)

        t = torch.randint(0, self.K_step, (B,), device=device).long()
        x = self._norm(y, self.norm_scale)
        x = x.transpose(1, 2)[:, None, :, :]  # [B, 1, M, T]

        noise = torch.randn_like(x)
        x_noisy = self.q_sample(x_start=x, t=t, noise=noise)
        x_recon = self.denoise_fn(x_noisy, t, cond)

        noise = noise.squeeze(1).transpose(1, 2)
        x_recon = x_recon.squeeze(1).transpose(1, 2)

        return noise, x_recon

    def inference(self, cond, lengths=None):
        B = cond.shape[0]
        device = cond.device

        if self.encoder is not None:
            cond = self.encoder(cond, lengths)

        # (B, M, T)
        cond = cond.transpose(1, 2)

        t = self.K_step
        shape = (cond.shape[0], 1, self.out_dim, cond.shape[2])
        x = torch.randn(shape, device=device)

        if self.pndm_speedup:
            self.noise_list = deque(maxlen=4)
            iteration_interval = int(self.pndm_speedup)
            for i in tqdm(
                reversed(range(0, t, iteration_interval)),
                desc="sample time step",
                total=t // iteration_interval,
            ):
                x = self.p_sample_plms(
                    x,
                    torch.full((B,), i, device=device, dtype=torch.long),
                    iteration_interval,
                    cond,
                )
        else:
            for i in tqdm(reversed(range(0, t)), desc="sample time step", total=t):
                x = self.p_sample(
                    x, torch.full((B,), i, device=device, dtype=torch.long), cond
                )
        x = self._denorm(x[:, 0].transpose(1, 2), self.norm_scale)
        return x



from pathlib import Path

import numpy as np
import pysptk
import pyworld
from nnmnkwii.frontend import merlin as fe
from nnmnkwii.io import hts
from nnmnkwii.preprocessing.f0 import interp1d
from nnsvs.gen import correct_vuv_by_phone, gen_spsvs_static_features
from scipy.io import wavfile


def _extract_static_feats(wav, sr):
    f0, timeaxis = pyworld.dio(wav, sr, frame_period=5)
    spectrogram = pyworld.cheaptrick(wav, f0, timeaxis, sr)
    aperiodicity = pyworld.d4c(wav, f0, timeaxis, sr)

    mgc = pysptk.sp2mc(spectrogram, order=59, alpha=pysptk.util.mcepalpha(sr))
    f0 = f0[:, None]
    lf0 = f0.copy()
    nonzero_indices = np.nonzero(f0)
    lf0[nonzero_indices] = np.log(f0[nonzero_indices])
    vuv = (lf0 != 0).astype(np.float32)
    lf0 = interp1d(lf0, kind="slinear")
    bap = pyworld.code_aperiodicity(aperiodicity, sr)

    feats = np.hstack((mgc, lf0, vuv, bap)).astype(np.float32)
    stream_sizes = [mgc.shape[1], lf0.shape[1], vuv.shape[1], bap.shape[1]]

    return feats, stream_sizes


def test_gen_spsvs_static_features():
    wav_path = Path(__file__).parent / "data" / "nitech_jp_song070_f001_004.wav"
    lab_path = Path(__file__).parent / "data" / "nitech_jp_song070_f001_004.lab"

    binary_dict, numeric_dict = hts.load_question_set(
        Path(__file__).parent / "data" / "jp_test.hed"
    )

    labels = hts.load(lab_path)
    sr, wav = wavfile.read(wav_path)
    wav = wav.astype(np.float64)
    assert sr == 48000

    out_feats, stream_sizes = _extract_static_feats(wav, sr)
    has_dynamic_features = [False] * len(stream_sizes)
    pitch_idx = len(binary_dict) + 1

    params = {
        "labels": labels,
        "acoustic_features": out_feats,
        "binary_dict": binary_dict,
        "numeric_dict": numeric_dict,
        "stream_sizes": stream_sizes,
        "has_dynamic_features": has_dynamic_features,
        "pitch_idx": pitch_idx,
        "relative_f0": False,
        "frame_period": 5,
        "force_fix_vuv": False,
    }

    mgc, lf0, vuv, bap = gen_spsvs_static_features(**params)
    assert mgc.shape[1] == 60
    assert lf0.shape[1] == 1
    assert vuv.shape[1] == 1
    assert bap.shape[1] == 5

    # w/o V/UV correction, vuv should't change
    out_vuv_idx = 61
    N = np.abs(vuv - out_feats[:, out_vuv_idx : out_vuv_idx + 1]).sum()
    assert int(N) == 0


def test_correct_vuv_by_phone():
    wav_path = Path(__file__).parent / "data" / "nitech_jp_song070_f001_004.wav"
    lab_path = Path(__file__).parent / "data" / "nitech_jp_song070_f001_004.lab"

    binary_dict, numeric_dict = hts.load_question_set(
        Path(__file__).parent / "data" / "jp_test.hed"
    )

    labels = hts.load(lab_path)
    sr, wav = wavfile.read(wav_path)
    wav = wav.astype(np.float64)
    assert sr == 48000

    out_feats, stream_sizes = _extract_static_feats(wav, sr)
    has_dynamic_features = [False] * len(stream_sizes)
    pitch_idx = len(binary_dict) + 1

    linguistic_features = fe.linguistic_features(
        labels,
        binary_dict,
        numeric_dict,
        add_frame_features=True,
        subphone_features="coarse_coding",
    )

    params = {
        "labels": labels,
        "acoustic_features": out_feats,
        "binary_dict": binary_dict,
        "numeric_dict": numeric_dict,
        "stream_sizes": stream_sizes,
        "has_dynamic_features": has_dynamic_features,
        "pitch_idx": pitch_idx,
        "relative_f0": False,
        "frame_period": 5,
    }

    out_vuv_idx = 61
    vuv = out_feats[:, out_vuv_idx : out_vuv_idx + 1]

    vuv_corrected = correct_vuv_by_phone(vuv, binary_dict, linguistic_features)
    # by correcting VUV should make a difference
    _, _, vuv_fixed, _ = gen_spsvs_static_features(**{**params, "force_fix_vuv": True})
    assert np.any(vuv_corrected != vuv)

    # 0: Rest 1: Voiced 2: Unvoiced
    rest_idx = 0
    voiced_idx = 1
    unvoiced_idx = 2
    assert np.all(vuv_corrected[linguistic_features[:, rest_idx] > 0] < 0.5)
    assert np.all(vuv_corrected[linguistic_features[:, voiced_idx] > 0] > 0.5)
    assert np.all(vuv_corrected[linguistic_features[:, unvoiced_idx] > 0] < 0.5)



# Utils for multi-stream features

import numpy as np
import torch
from nnmnkwii import paramgen


def get_windows(num_window=1):
    """Get windows for MLPG.

    Args:
        num_window (int): number of windows

    Returns:
        list: list of windows
    """
    windows = [(0, 0, np.array([1.0]))]
    if num_window >= 2:
        windows.append((1, 1, np.array([-0.5, 0.0, 0.5])))
    if num_window >= 3:
        windows.append((1, 1, np.array([1.0, -2.0, 1.0])))

    if num_window >= 4:
        raise ValueError(f"Not supported num windows: {num_window}")

    return windows


def select_streams(
    inputs,
    stream_sizes=None,
    streams=None,
    concat=True,
):
    """Select streams from multi-stream features

    Args:
        inputs (array like): input 3-d or 2-d array
        stream_sizes (list): stream sizes
        streams (list): Streams of interests. Returns all streams if streams is None.
        concat (bool): Concatenate streams. Defaults to True.

    Returns:
        array like: selected streams
    """
    if stream_sizes is None:
        stream_sizes = [60, 1, 1, 1]
    if streams is None:
        streams = [True] * len(stream_sizes)
    ret = []
    start_indices = np.hstack(([0], np.cumsum(stream_sizes)[:-1]))
    for start_idx, size, enabled in zip(start_indices, stream_sizes, streams):
        if not enabled:
            continue
        if len(inputs.shape) == 3:
            s = inputs[:, :, start_idx : start_idx + size]
        else:
            s = inputs[:, start_idx : start_idx + size]
        ret.append(s)

    if not concat:
        return ret

    if isinstance(inputs, torch.Tensor):
        return torch.cat(ret, dim=-1)
    else:
        return np.concatenate(ret, -1)


def split_streams(inputs, stream_sizes=None):
    """Split streams from multi-stream features

    Args:
        inputs (array like): input 3-d array
        stream_sizes (list): sizes for each stream

    Returns:
        list: list of stream features
    """
    if stream_sizes is None:
        stream_sizes = [60, 1, 1, 1]
    ret = []
    start_indices = np.hstack(([0], np.cumsum(stream_sizes)[:-1]))
    for start_idx, size in zip(start_indices, stream_sizes):
        if len(inputs.shape) == 3:
            s = inputs[:, :, start_idx : start_idx + size]
        else:
            s = inputs[:, start_idx : start_idx + size]
        ret.append(s)

    return ret


def get_static_stream_sizes(stream_sizes, has_dynamic_features, num_windows):
    """Get stream sizes for static features

    Args:
        inputs (array like): input 3-d or 2-d array
        num_windows (int): number of windows
        stream_sizes (list): stream sizes
        has_dynamic_features (list): binary flags that indicates if steams have dynamic features
        streams (list, optional): Streams of interests. Returns all streams if streams is None.
            Defaults to None.

    Returns:
        list: stream sizes
    """
    static_stream_sizes = np.array(stream_sizes)
    static_stream_sizes[has_dynamic_features] = (
        static_stream_sizes[has_dynamic_features] / num_windows
    )

    return static_stream_sizes


def get_static_features(
    inputs,
    num_windows,
    stream_sizes=None,
    has_dynamic_features=None,
    streams=None,
):
    """Get static features from static+dynamic features

    Args:
        inputs (array like): input 3-d or 2-d array
        num_windows (int): number of windows
        stream_sizes (list): stream sizes
        has_dynamic_features (list): binary flags that indicates if steams have dynamic features
        streams (list, optional): Streams of interests. Returns all streams if streams is None.
            Defaults to None.

    Returns:
        list: list of static features
    """
    if stream_sizes is None:
        stream_sizes = [180, 3, 1, 15]
    if has_dynamic_features is None:
        has_dynamic_features = [True, True, False, True]
    if streams is None:
        streams = [True] * len(stream_sizes)
    _, _, D = inputs.shape
    if stream_sizes is None or (len(stream_sizes) == 1 and has_dynamic_features[0]):
        return inputs[:, :, : D // num_windows]
    if len(stream_sizes) == 1 and not has_dynamic_features[0]:
        return inputs

    # Multi stream case
    ret = []
    start_indices = np.hstack(([0], np.cumsum(stream_sizes)[:-1]))
    for start_idx, size, v, enabled in zip(
        start_indices, stream_sizes, has_dynamic_features, streams
    ):
        start_idx = int(start_idx)
        size = int(size)
        if not enabled:
            continue
        if v:
            static_features = inputs[:, :, start_idx : start_idx + size // num_windows]
        else:
            static_features = inputs[:, :, start_idx : start_idx + size]
        ret.append(static_features)
    return ret


def multi_stream_mlpg(
    inputs,
    variances,
    windows,
    stream_sizes=None,
    has_dynamic_features=None,
    streams=None,
):
    """Split streams and do apply MLPG if stream has dynamic features

    Args:
        inputs (array like): input 3-d or 2-d array
        variances (array like): variances of input features
        windows (list): windows for parameter generation
        stream_sizes (list): stream sizes
        has_dynamic_features (list): binary flags that indicates if steams have dynamic features
        streams (list, optional): Streams of interests. Returns all streams if streams is None.
            Defaults to None.

    Raises:
        RuntimeError: if stream sizes are wrong

    Returns:
        array like: generated static features
    """
    if stream_sizes is None:
        stream_sizes = [180, 3, 1, 3]
    if has_dynamic_features is None:
        has_dynamic_features = [True, True, False, True]
    if streams is None:
        streams = [True] * len(stream_sizes)
    T, D = inputs.shape
    if D != sum(stream_sizes):
        raise RuntimeError("You probably have specified wrong dimension params.")

    # Straem indices for static+delta features
    # [0,   180, 183, 184]
    start_indices = np.hstack(([0], np.cumsum(stream_sizes)[:-1]))
    # [180, 183, 184, 199]
    end_indices = np.cumsum(stream_sizes)

    ret = []
    for in_start_idx, in_end_idx, v, enabled in zip(
        start_indices,
        end_indices,
        has_dynamic_features,
        streams,
    ):
        if not enabled:
            continue
        x = inputs[:, in_start_idx:in_end_idx]
        if inputs.shape == variances.shape:
            var_ = variances[:, in_start_idx:in_end_idx]
        else:
            var_ = np.tile(variances[in_start_idx:in_end_idx], (T, 1))
        y = paramgen.mlpg(x, var_, windows) if v else x
        ret.append(y)

    return np.concatenate(ret, -1)



import torch
import torch.nn.functional as F
from torch import nn


class MDNLayer(nn.Module):
    """Mixture Density Network layer

    The input maps to the parameters of a Mixture of Gaussians (MoG) probability
    distribution, where each Gaussian has out_dim dimensions and diagonal covariance.
    If dim_wise is True, features for each dimension are modeld by independent 1-D GMMs
    instead of modeling jointly. This would workaround training difficulty
    especially for high dimensional data.

    Implementation references:
    1. Mixture Density Networks by Mike Dusenberry
    https://mikedusenberry.com/mixture-density-networks
    2. PRML book
    https://www.microsoft.com/en-us/research/people/cmbishop/prml-book/
    3. sagelywizard/pytorch-mdn
    https://github.com/sagelywizard/pytorch-mdn
    4. sksq96/pytorch-mdn
    https://github.com/sksq96/pytorch-mdn

    Attributes:
        in_dim (int): the number of dimensions in the input
        out_dim (int): the number of dimensions in the output
        num_gaussians (int): the number of mixture component
        dim_wise (bool): whether to model data for each dimension separately
    """

    def __init__(self, in_dim, out_dim, num_gaussians=30, dim_wise=False):
        super(MDNLayer, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.num_gaussians = num_gaussians
        self.dim_wise = dim_wise

        odim_log_pi = out_dim * num_gaussians if dim_wise else num_gaussians
        self.log_pi = nn.Linear(in_dim, odim_log_pi)

        self.log_sigma = nn.Linear(in_dim, out_dim * num_gaussians)
        self.mu = nn.Linear(in_dim, out_dim * num_gaussians)

    def forward(self, minibatch):
        """Forward for MDN

        Args:
            minibatch (torch.Tensor): tensor of shape (B, T, D_in)
                B is the batch size and T is data lengths of this batch,
                and D_in is in_dim.

        Returns:
            torch.Tensor: Tensor of shape (B, T, G) or (B, T, G, D_out)
                Log of mixture weights. G is num_gaussians and D_out is out_dim.
            torch.Tensor: Tensor of shape (B, T, G, D_out)
                the log of standard deviation of each Gaussians.
            torch.Tensor: Tensor of shape (B, T, G, D_out)
                mean of each Gaussians
        """
        B = len(minibatch)
        if self.dim_wise:
            # (B, T, G, D_out)
            log_pi = self.log_pi(minibatch).view(
                B, -1, self.num_gaussians, self.out_dim
            )
            log_pi = F.log_softmax(log_pi, dim=2)
        else:
            # (B, T, G)
            log_pi = F.log_softmax(self.log_pi(minibatch), dim=2)
        log_sigma = self.log_sigma(minibatch)
        log_sigma = log_sigma.view(B, -1, self.num_gaussians, self.out_dim)
        mu = self.mu(minibatch)
        mu = mu.view(B, -1, self.num_gaussians, self.out_dim)
        return log_pi, log_sigma, mu


def mdn_loss(
    log_pi, log_sigma, mu, target, log_pi_min=-7.0, log_sigma_min=-7.0, reduce=True
):
    """Calculates the error, given the MoG parameters and the target.
    The loss is the negative log likelihood of the data given the MoG
    parameters.

    Args:
        log_pi (torch.Tensor): Tensor of shape (B, T, G) or (B, T, G, D_out)
            The log of multinomial distribution of the Gaussians. B is the batch size,
            T is data length of this batch, and G is num_gaussians of class MDNLayer.
        log_sigma (torch.Tensor): Tensor of shape (B, T, G ,D_out)
            The log standard deviation of the Gaussians. D_out is out_dim of class
            MDNLayer.
        mu (torch.Tensor): Tensor of shape (B, T, G, D_out)
            The means of the Gaussians.
        target (torch.Tensor): Tensor of shape (B, T, D_out)
            The target variables.
        log_pi_min (float): Minimum value of log_pi (for numerical stability)
        log_sigma_min (float): Minimum value of log_sigma (for numerical stability)
        reduce: If True, the losses are averaged for each batch.

    Returns:
        loss (B) or (B, T): Negative Log Likelihood of Mixture Density Networks.
    """
    dim_wise = len(log_pi.shape) == 4

    # Clip log_sigma and log_pi with log_clamp_min for numerical stability
    log_sigma = torch.clamp(log_sigma, min=log_sigma_min)
    log_pi = torch.clamp(log_pi, min=log_pi_min)

    # Expand the dim of target as (B, T, D_out) -> (B, T, 1, D_out) -> (B, T,G, D_out)
    target = target.unsqueeze(2).expand_as(log_sigma)

    # Center target variables and clamp them within +/- 5SD for numerical stability.
    centered_target = target - mu
    scale = torch.exp(log_sigma)
    edge = 5 * scale
    centered_target = torch.where(centered_target > edge, edge, centered_target)
    centered_target = torch.where(centered_target < -edge, -edge, centered_target)

    # Create gaussians with mean=0 and variance=torch.exp(log_sigma)^2
    dist = torch.distributions.Normal(loc=0, scale=scale)

    log_prob = dist.log_prob(centered_target)

    if dim_wise:
        # (B, T, D_out. D_out)
        loss = log_prob + log_pi
    else:
        # Here we assume that the covariance matrix of multivariate Gaussian
        # distribution is diagonal to handle the mean and the variance in each
        # dimension separately.
        # Reference:
        # https://markusthill.github.io/gaussian-distribution-with-a-diagonal-covariance-matrix/
        # log pi(x)N(y|mu(x),sigma(x)) = log pi(x) + log N(y|mu(x),sigma(x))
        # log N(y_1,y_2,...,y_{D_out}|mu(x),sigma(x))
        #  = log N(y_1|mu(x),sigma(x))...N(y_{D_out}|mu(x),sigma(x))
        #  = \sum_{i=1}^{D_out} log N(y_i|mu(x),sigma(x))
        # (B, T, G, D_out) -> (B, T, G)
        loss = torch.sum(log_prob, dim=3) + log_pi

    # Calculate negative log likelihood.
    # Use torch.log_sum_exp instead of the combination of torch.sum and torch.log
    # (Reference: https://github.com/r9y9/nnsvs/pull/20#discussion_r495514563)
    # if dim_wise is True: (B, T, G, D_out) -> (B, T, D_out)
    # else (B, T, G) -> (B, T)
    loss = -torch.logsumexp(loss, dim=2)

    if reduce:
        # (B, T) -> (B)
        return torch.mean(loss, dim=1)
    else:
        # not averaged (for applying mask later)
        # (B, T)
        return loss
    return


# from r9y9/wavenet_vocoder/wavenet_vocoder/mixture.py
def to_one_hot(tensor, n, fill_with=1.0):
    # we perform one hot encore with respect to the last axis
    one_hot = torch.FloatTensor(tensor.size() + (n,)).zero_()
    if tensor.is_cuda:
        one_hot = one_hot.cuda()
    one_hot.scatter_(len(tensor.size()), tensor.unsqueeze(-1), fill_with)
    return one_hot


def mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu):
    """Return the mean and standard deviation of the Gaussian component
    whose weight coefficient is the largest as the most probable predictions.

    Args:
        log_pi (torch.Tensor): Tensor of shape (B, T, G) or (B, T, G, D_out)
            The log of multinomial distribution of the Gaussians.
            B is the batch size, T is data length of this batch,
            G is num_gaussians of class MDNLayer.
        log_sigma (torch.Tensor): Tensor of shape (B, T, G, D_out)
            The standard deviation of the Gaussians. D_out is out_dim of class
            MDNLayer.
        mu (torch.Tensor): Tensor of shape (B, T, G, D_out)
            The means of the Gaussians. D_out is out_dim of class MDNLayer.

    Returns:
        tuple: tuple of torch.Tensor
            torch.Tensor of shape (B, T, D_out). The standardd deviations
            of the most probable Gaussian component.
            torch.Tensor of shape (B, T, D_out). Means of the Gaussians.
    """
    dim_wise = len(log_pi.shape) == 4
    _, _, num_gaussians, _ = mu.shape
    # Get the indexes of the largest log_pi
    _, max_component = torch.max(log_pi, dim=2)  # (B, T) or (B, T, C_out)

    # Convert max_component to one_hot manner
    # if dim_wise: (B, T, D_out) -> (B, T, D_out, G)
    # else: (B, T) -> (B, T, G)
    one_hot = to_one_hot(max_component, num_gaussians)

    if dim_wise:
        # (B, T, G, D_out)
        one_hot = one_hot.transpose(2, 3)
        assert one_hot.shape == mu.shape
    else:
        # Expand the dim of one_hot as  (B, T, G) -> (B, T, G, d_out)
        one_hot = one_hot.unsqueeze(3).expand_as(mu)

    # Multiply one_hot and sum to get mean(mu) and standard deviation(sigma)
    # of the Gaussians whose weight coefficient(log_pi) is the largest.
    #  (B, T, G, d_out) -> (B, T, d_out)
    max_mu = torch.sum(mu * one_hot, dim=2)
    max_sigma = torch.exp(torch.sum(log_sigma * one_hot, dim=2))

    return max_sigma, max_mu


def mdn_get_sample(log_pi, log_sigma, mu):
    """Sample from mixture of the Gaussian component whose weight coefficient is
    the largest as the most probable predictions.

    Args:
        log_pi (torch.Tensor): Tensor of shape (B, T, G) or (B, T, G, D_out)
            The log of multinomial distribution of the Gaussians.
            B is the batch size, T is data length of this batch,
            G is num_gaussians of class MDNLayer.
        log_sigma (torch.Tensor): Tensor of shape (B, T, G, D_out)
            The log of standard deviation of the Gaussians.
            D_out is out_dim of class MDNLayer.
        mu (torch.Tensor): Tensor of shape (B, T, G, D_out)
            The means of the Gaussians. D_out is out_dim of class MDNLayer.

    Returns:
        torch.Tensor: Tensor of shape (B, T, D_out)
            Sample from the mixture of the Gaussian component.
    """
    max_sigma, max_mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)

    # Create gaussians with mean=max_mu and variance=max_log_sigma^2
    dist = torch.distributions.Normal(loc=max_mu, scale=max_sigma)

    # Sample from normal distribution
    sample = dist.sample()

    return sample



import torch
from scipy import signal
from torch import nn
from torch.nn import functional as F

# Part of code was adapted from:
# https://github.com/nii-yamagishilab/project-NN-Pytorch-scripts


def lowpass_filter(x, fs, cutoff=5, N=5):
    """Lowpass filter

    Args:
        x (np.ndarray): input signal
        fs (int): sampling rate
        cutoff (int): cutoff frequency

    Returns:
        np.ndarray: filtered signal
    """
    nyquist = fs // 2
    norm_cutoff = cutoff / nyquist
    Wn = [norm_cutoff]

    b, a = signal.butter(N, Wn, "lowpass")
    if len(x) <= max(len(a), len(b)) * (N // 2 + 1):
        # NOTE: input signal is too short
        return x

    # NOTE: use zero-phase filter
    y = signal.filtfilt(b, a, x)

    return y


def bandpass_filter(x, sr, cutoff=70):
    """Band-pass filter

    Args:
        x (np.ndarray): input signal
        fs (int): sampling rate
        cutoff (int): cutoff frequency

    Returns:
        np.ndarray: filtered signal
    """
    nyquist = sr // 2
    norm_cutoff = cutoff / nyquist
    Wn = [norm_cutoff, 0.999]

    b, a = signal.butter(5, Wn, "bandpass")
    y = signal.filtfilt(b, a, x)

    return y


class TimeInvFIRFilter(nn.Conv1d):
    """Time-invatiant FIR filter implementation

    Args:
        channels (int): input channels
        filt_coef (torch.Tensor): FIR filter coefficients
        causal (bool): causal
        requires_grad (bool): trainable kernel or not
    """

    def __init__(self, channels, filt_coef, causal=True, requires_grad=False):
        # assuming 1-D filter coef vector and odd num taps
        assert len(filt_coef.shape) == 1
        # assert len(filt_coef) % 2 == 1
        kernel_size = len(filt_coef)
        self.causal = causal
        if causal:
            padding = (kernel_size - 1) * 1
        else:
            padding = (kernel_size - 1) // 2 * 1
        # channel-wise filtering (groups=channels)
        super(TimeInvFIRFilter, self).__init__(
            channels, channels, kernel_size, padding=padding, groups=channels, bias=None
        )
        self.weight.data[:, :, :] = filt_coef.flip(-1)
        self.weight.requires_grad = requires_grad

    def forward(self, x):
        out = super(TimeInvFIRFilter, self).forward(x)
        out = out[:, :, : -self.padding[0]] if self.causal else out
        return out


class TrTimeInvFIRFilter(nn.Conv1d):
    """Trainable Time-invatiant FIR filter implementation

    H(z) = \\sigma_{k=0}^{filt_dim} b_{k}z_{-k}

    Note that b_{0} is fixed to 1 if fixed_0th is True.

    Args:
        channels (int): input channels
        filt_dim (int): FIR filter dimension
        causal (bool): causal
        tanh (bool): apply tanh to filter coef or not.
        fixed_0th (bool): fix the first filt coef to 1 or not.
    """

    def __init__(self, channels, filt_dim, causal=True, tanh=True, fixed_0th=True):
        # Initialize filt coef with small random values
        init_filt_coef = torch.randn(filt_dim) * (1 / filt_dim)
        # assert len(filt_coef) % 2 == 1
        kernel_size = len(init_filt_coef)
        self.causal = causal
        if causal:
            padding = (kernel_size - 1) * 1
        else:
            padding = (kernel_size - 1) // 2 * 1
        # channel-wise filtering (groups=channels)
        super(TrTimeInvFIRFilter, self).__init__(
            channels, channels, kernel_size, padding=padding, groups=channels, bias=None
        )
        self.weight.data[:, :, :] = init_filt_coef.flip(-1)
        self.weight.requires_grad = True
        self.tanh = tanh
        self.fixed_0th = fixed_0th

    def get_filt_coefs(self):
        # apply tanh for filtter stability
        b = torch.tanh(self.weight) if self.tanh else self.weight
        b = b.clone()
        if self.fixed_0th:
            b[:, :, -1] = 1
        return b

    def forward(self, x):
        b = self.get_filt_coefs()
        out = F.conv1d(
            x, b, self.bias, self.stride, self.padding, self.dilation, self.groups
        )
        if self.padding[0] > 0:
            out = out[:, :, : -self.padding[0]] if self.causal else out
        return out



import numpy as np
import pysinsy
import pytest
from nnmnkwii.io import hts
from nnsvs.io.hts import full_to_mono
from nnsvs.pretrained import retrieve_pretrained_model
from nnsvs.svs import NEUTRINO, SPSVS
from nnsvs.util import example_xml_file


@pytest.mark.parametrize("post_filter_type", ["merlin", "gv", "nnsvs", "none"])
@pytest.mark.parametrize("vocoder_type", ["world"])
def test_svs(post_filter_type, vocoder_type):
    model_dir = retrieve_pretrained_model("r9y9/yoko_latest")
    engine = SPSVS(model_dir, verbose=100)

    contexts = pysinsy.extract_fullcontext(example_xml_file(key="get_over"))
    labels = hts.HTSLabelFile.create_from_contexts(contexts)

    wav, sr = engine.svs(
        labels,
        post_filter_type=post_filter_type,
        vocoder_type=vocoder_type,
    )
    assert sr == 48000
    assert np.isfinite(wav).all()


def test_svs_options():
    model_dir = retrieve_pretrained_model("r9y9/yoko_latest")
    engine = SPSVS(model_dir, verbose=100)

    contexts = pysinsy.extract_fullcontext(example_xml_file(key="get_over"))
    labels = hts.HTSLabelFile.create_from_contexts(contexts)

    engine.svs(
        labels,
        fill_silence_to_rest=True,
    )
    engine.svs(
        labels,
        force_fix_vuv=True,
    )


def test_segmented_svs():
    model_dir = retrieve_pretrained_model("r9y9/yoko_latest")
    engine = SPSVS(model_dir, verbose=100)

    contexts = pysinsy.extract_fullcontext(example_xml_file(key="get_over"))
    labels = hts.HTSLabelFile.create_from_contexts(contexts)

    wav, sr = engine.svs(
        labels,
        segmented_synthesis=True,
    )
    assert sr == 48000
    assert np.isfinite(wav).all()


def test_neutrino():
    model_dir = retrieve_pretrained_model("r9y9/yoko_latest")
    engine = NEUTRINO(model_dir, verbose=100)

    musicxml = example_xml_file(key="get_over")
    full_labels, mono_labels = engine.musicxml2label(musicxml)

    assert len(full_labels) == len(mono_labels)

    num_phrases = engine.get_num_phrases(mono_labels)
    assert num_phrases >= 1

    timing_labels = full_to_mono(engine.predict_timing(full_labels))
    f0, mgc, bap = engine.predict_acoustic(full_labels, timing_labels)
    assert len(f0) == len(mgc) == len(bap)

    wav = engine.predict_waveform(f0, mgc, bap, vocoder_type="world")
    assert np.isfinite(wav).all()

    wav, sr = engine.svs(full_labels)
    assert np.isfinite(wav).all()
    assert sr == 48000


def test_neutrino_phrase():
    model_dir = retrieve_pretrained_model("r9y9/yoko_latest")
    engine = NEUTRINO(model_dir, verbose=100)

    musicxml = example_xml_file(key="get_over")
    full_labels, mono_labels = engine.musicxml2label(musicxml)
    num_phrases = engine.get_num_phrases(mono_labels)
    assert num_phrases >= 1

    timing_labels = full_to_mono(engine.predict_timing(full_labels))

    phraselist = engine.get_phraselist(full_labels, timing_labels)
    print(phraselist)

    f0, mgc, bap = engine.predict_acoustic(full_labels, timing_labels, phrase_num=1)
    assert len(f0) == len(mgc) == len(bap)

    wav = engine.predict_waveform(f0, mgc, bap, vocoder_type="world")
    assert np.isfinite(wav).all()



# coding: utf-8

import os
from os.path import basename, join

import hydra
import joblib
import numpy as np
import torch
from hydra.utils import to_absolute_path
from nnmnkwii.datasets import FileSourceDataset
from nnsvs.base import PredictionType
from nnsvs.logger import getLogger
from nnsvs.multistream import get_windows, multi_stream_mlpg
from nnsvs.train_util import NpyFileSource
from omegaconf import DictConfig, OmegaConf
from tqdm import tqdm

logger = None

use_cuda = torch.cuda.is_available()


@hydra.main(config_path="conf/generate", config_name="config")
def my_app(config: DictConfig) -> None:
    global logger
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    device = torch.device("cuda" if use_cuda else "cpu")
    in_dir = to_absolute_path(config.in_dir)
    out_dir = to_absolute_path(config.out_dir)
    os.makedirs(out_dir, exist_ok=True)

    model_config = OmegaConf.load(to_absolute_path(config.model.model_yaml))
    model = hydra.utils.instantiate(model_config.netG).to(device)
    checkpoint = torch.load(
        to_absolute_path(config.model.checkpoint),
        map_location=lambda storage, loc: storage,
    )
    model.load_state_dict(checkpoint["state_dict"])
    model.eval()

    scaler = joblib.load(to_absolute_path(config.out_scaler_path))

    in_feats = FileSourceDataset(NpyFileSource(in_dir, logger))

    with torch.no_grad():
        for idx in tqdm(range(len(in_feats))):
            feats = torch.from_numpy(in_feats[idx]).unsqueeze(0).to(device)

            if model.prediction_type() == PredictionType.PROBABILISTIC:

                max_mu, max_sigma = model.inference(feats, [feats.shape[1]])

                if np.any(model_config.has_dynamic_features):
                    # Apply denormalization
                    # (B, T, D_out) -> (T, D_out)
                    max_sigma_sq = (
                        max_sigma.squeeze(0).cpu().data.numpy() ** 2 * scaler.var_
                    )
                    max_mu = scaler.inverse_transform(
                        max_mu.squeeze(0).cpu().data.numpy()
                    )
                    # Apply MLPG
                    # (T, D_out) -> (T, static_dim)
                    out = multi_stream_mlpg(
                        max_mu,
                        max_sigma_sq,
                        get_windows(model_config.num_windows),
                        model_config.stream_sizes,
                        model_config.has_dynamic_features,
                    )

                else:
                    # (T, D_out)
                    out = max_mu.squeeze(0).cpu().data.numpy()
                    out = scaler.inverse_transform(out)
            else:
                out = (
                    model.inference(feats, [feats.shape[1]])
                    .squeeze(0)
                    .cpu()
                    .data.numpy()
                )
                out = scaler.inverse_transform(out)

                # Apply MLPG if necessary
                if np.any(model_config.has_dynamic_features):
                    out = multi_stream_mlpg(
                        out,
                        scaler.var_,
                        get_windows(model_config.num_windows),
                        model_config.stream_sizes,
                        model_config.has_dynamic_features,
                    )

            name = basename(in_feats.collected_files[idx][0])
            out_path = join(out_dir, name)
            np.save(out_path, out, allow_pickle=False)


def entry():
    my_app()


if __name__ == "__main__":
    my_app()



# coding: utf-8
import os
from concurrent.futures import ProcessPoolExecutor
from os.path import basename, join, splitext

import hydra
import numpy as np
from hydra.utils import to_absolute_path
from nnmnkwii.datasets import FileSourceDataset
from nnsvs.data import (
    DurationFeatureSource,
    MelF0AcousticSource,
    MusicalLinguisticSource,
    TimeLagFeatureSource,
    WORLDAcousticSource,
)
from nnsvs.logger import getLogger
from omegaconf import DictConfig, OmegaConf
from tqdm import tqdm

logger = None


def _prepare_timelag_feature(
    in_timelag_root,
    out_timelag_root,
    in_timelag: FileSourceDataset,
    out_timelag: FileSourceDataset,
    idx: int,
) -> None:
    """prepare timelag feature for one item of in_duration"""
    x, y = in_timelag[idx], out_timelag[idx]
    name = splitext(basename(in_timelag.collected_files[idx][0]))[0]
    xpath = join(in_timelag_root, name + "-feats.npy")
    ypath = join(out_timelag_root, name + "-feats.npy")
    np.save(xpath, x, allow_pickle=False)
    np.save(ypath, y, allow_pickle=False)


def _prepare_duration_feature(
    in_duration_root,
    out_duration_root,
    in_duration: FileSourceDataset,
    out_duration: FileSourceDataset,
    idx: int,
) -> None:
    """prepare duration feature for one item of in_duration"""
    x, y = in_duration[idx], out_duration[idx]
    name = splitext(basename(in_duration.collected_files[idx][0]))[0]
    xpath = join(in_duration_root, name + "-feats.npy")
    ypath = join(out_duration_root, name + "-feats.npy")
    np.save(xpath, x, allow_pickle=False)
    np.save(ypath, y, allow_pickle=False)


def _prepare_acoustic_feature(
    in_acoustic_root,
    out_acoustic_root,
    out_postfilter_root,
    in_acoustic: FileSourceDataset,
    out_acoustic: FileSourceDataset,
    idx: int,
) -> None:
    """prepare acoustic feature for one item of in_acoustic"""
    x, (y, wave, y_pf) = in_acoustic[idx], out_acoustic[idx]
    name = splitext(basename(in_acoustic.collected_files[idx][0]))[0]
    if y is None:
        print(f"{name} is skipped")
        return
    xpath = join(in_acoustic_root, name + "-feats.npy")
    ypath = join(out_acoustic_root, name + "-feats.npy")
    wpath = join(out_acoustic_root, name + "-wave.npy")
    pfpath = join(out_postfilter_root, name + "-feats.npy")
    np.save(xpath, x, allow_pickle=False)
    np.save(ypath, y, allow_pickle=False)
    np.save(wpath, wave, allow_pickle=False)
    np.save(pfpath, y_pf, allow_pickle=False)


@hydra.main(config_path="conf/prepare_features", config_name="config")
def my_app(config: DictConfig) -> None:
    global logger
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    utt_list = to_absolute_path(config.utt_list)
    out_dir = to_absolute_path(config.out_dir)
    question_path_general = to_absolute_path(config.question_path)

    # Time-lag model
    # in: musical/linguistic context
    # out: time-lag (i.e. onset time deviation)
    if config.timelag.question_path is not None:
        question_path = config.timelag.question_path
    else:
        question_path = question_path_general

    in_timelag_source = MusicalLinguisticSource(
        utt_list,
        to_absolute_path(config.timelag.label_phone_score_dir),
        add_frame_features=False,
        subphone_features=None,
        question_path=question_path,
        log_f0_conditioning=config.log_f0_conditioning,
    )
    out_timelag_source = TimeLagFeatureSource(
        utt_list,
        to_absolute_path(config.timelag.label_phone_score_dir),
        to_absolute_path(config.timelag.label_phone_align_dir),
    )

    in_timelag = FileSourceDataset(in_timelag_source)
    out_timelag = FileSourceDataset(out_timelag_source)

    # Duration model
    # in: musical/linguistic context
    # out: duration
    if config.duration.question_path is not None:
        question_path = config.duration.question_path
    else:
        question_path = question_path_general

    in_duration_source = MusicalLinguisticSource(
        utt_list,
        to_absolute_path(config.duration.label_dir),
        add_frame_features=False,
        subphone_features=None,
        question_path=question_path,
        log_f0_conditioning=config.log_f0_conditioning,
    )
    out_duration_source = DurationFeatureSource(
        utt_list, to_absolute_path(config.duration.label_dir)
    )

    in_duration = FileSourceDataset(in_duration_source)
    out_duration = FileSourceDataset(out_duration_source)

    # Acoustic model
    # in: musical/linguistic context
    # out: acoustic features
    if config.acoustic.question_path is not None:
        question_path = config.acoustic.question_path
    else:
        question_path = question_path_general
    in_acoustic_source = MusicalLinguisticSource(
        utt_list,
        to_absolute_path(config.acoustic.label_dir),
        question_path,
        add_frame_features=True,
        subphone_features=config.acoustic.subphone_features,
        log_f0_conditioning=config.log_f0_conditioning,
        frame_period=config.acoustic.frame_period,
    )
    if config.acoustic.feature_type == "world":
        out_acoustic_source = WORLDAcousticSource(
            utt_list,
            to_absolute_path(config.acoustic.wav_dir),
            to_absolute_path(config.acoustic.label_dir),
            question_path,
            f0_extractor=config.acoustic.f0_extractor,
            f0_ceil=config.acoustic.f0_ceil,
            f0_floor=config.acoustic.f0_floor,
            frame_period=config.acoustic.frame_period,
            mgc_order=config.acoustic.mgc_order,
            num_windows=config.acoustic.num_windows,
            relative_f0=config.acoustic.relative_f0,
            vibrato_mode=config.acoustic.vibrato_mode,
            sample_rate=config.acoustic.sample_rate,
            d4c_threshold=config.acoustic.d4c_threshold,
            trajectory_smoothing=config.acoustic.trajectory_smoothing,
            trajectory_smoothing_cutoff=config.acoustic.trajectory_smoothing_cutoff,
            trajectory_smoothing_f0=config.acoustic.trajectory_smoothing_f0,
            trajectory_smoothing_cutoff_f0=config.acoustic.trajectory_smoothing_cutoff_f0,
            correct_vuv=config.acoustic.correct_vuv,
            correct_f0=config.acoustic.correct_f0,
            dynamic_features_flags=config.acoustic.dynamic_features_flags,
            use_world_codec=config.acoustic.use_world_codec,
            res_type=config.acoustic.res_type,
            use_mcep_aperiodicity=config.acoustic.use_mcep_aperiodicity,
            mcep_aperiodicity_order=config.acoustic.mcep_aperiodicity_order,
        )
    elif config.acoustic.feature_type == "melf0":
        out_acoustic_source = MelF0AcousticSource(
            utt_list,
            to_absolute_path(config.acoustic.wav_dir),
            to_absolute_path(config.acoustic.label_dir),
            question_path,
            f0_extractor=config.acoustic.f0_extractor,
            f0_ceil=config.acoustic.f0_ceil,
            f0_floor=config.acoustic.f0_floor,
            frame_period=config.acoustic.frame_period,
            sample_rate=config.acoustic.sample_rate,
            d4c_threshold=config.acoustic.d4c_threshold,
            trajectory_smoothing_f0=config.acoustic.trajectory_smoothing_f0,
            trajectory_smoothing_cutoff_f0=config.acoustic.trajectory_smoothing_cutoff_f0,
            correct_vuv=config.acoustic.correct_vuv,
            correct_f0=config.acoustic.correct_f0,
            res_type=config.acoustic.res_type,
            fft_size=config.acoustic.fft_size,
            win_length=config.acoustic.win_length,
            hop_size=config.acoustic.hop_size,
            fmin=config.acoustic.fmin,
            fmax=config.acoustic.fmax,
            eps=config.acoustic.eps,
            num_mels=config.acoustic.num_mels,
        )
    else:
        raise ValueError(
            "Unknown feature type: {}".format(config.acoustic.feature_type)
        )

    in_acoustic = FileSourceDataset(in_acoustic_source)
    out_acoustic = FileSourceDataset(out_acoustic_source)

    # Save as files
    in_timelag_root = join(out_dir, "in_timelag")
    out_timelag_root = join(out_dir, "out_timelag")
    in_duration_root = join(out_dir, "in_duration")
    out_duration_root = join(out_dir, "out_duration")
    in_acoustic_root = join(out_dir, "in_acoustic")
    out_acoustic_root = join(out_dir, "out_acoustic")
    out_postfilter_root = join(out_dir, "out_postfilter")

    for d in [
        in_timelag_root,
        out_timelag_root,
        in_duration_root,
        out_duration_root,
        in_acoustic_root,
        out_acoustic_root,
        out_postfilter_root,
    ]:
        if not os.path.exists(d):
            logger.info("mkdirs: %s", format(d))
            os.makedirs(d)

    # Save features for timelag model
    if config.timelag.enabled:
        logger.info("Timelag linguistic feature dim: %s", str(in_timelag[0].shape[1]))
        logger.info("Timelag feature dim: %s", str(out_timelag[0].shape[1]))
        with ProcessPoolExecutor(max_workers=config.max_workers) as executor:
            futures = [
                executor.submit(
                    _prepare_timelag_feature,
                    in_timelag_root,
                    out_timelag_root,
                    in_timelag,
                    out_timelag,
                    idx,
                )
                for idx in range(len(in_timelag))
            ]
            for future in tqdm(futures):
                future.result()

    # Save features for duration model
    if config.duration.enabled:
        logger.info("Duration linguistic feature dim: %s", str(in_duration[0].shape[1]))
        logger.info("Duration feature dim: %s", str(out_duration[0].shape[1]))
        with ProcessPoolExecutor(max_workers=config.max_workers) as executor:
            futures = [
                executor.submit(
                    _prepare_duration_feature,
                    in_duration_root,
                    out_duration_root,
                    in_duration,
                    out_duration,
                    idx,
                )
                for idx in range(len(in_duration))
            ]
            for future in tqdm(futures):
                future.result()

    # Save features for acoustic model
    if config.acoustic.enabled:
        logger.info("Acoustic linguistic feature dim: %s", str(in_acoustic[0].shape[1]))
        logger.info("Acoustic feature dim: %s", str(out_acoustic[0][0].shape[1]))
        with ProcessPoolExecutor(max_workers=config.max_workers) as executor:
            futures = [
                executor.submit(
                    _prepare_acoustic_feature,
                    in_acoustic_root,
                    out_acoustic_root,
                    out_postfilter_root,
                    in_acoustic,
                    out_acoustic,
                    idx,
                )
                for idx in range(len(in_acoustic))
            ]
            for future in tqdm(futures):
                future.result()


def entry():
    my_app()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



from functools import partial

from nnsvs.acoustic_models.multistream import (
    MDNMultistreamSeparateF0MelModel,
    MultistreamSeparateF0MelModel,
    MultistreamSeparateF0ParametricModel,
    NPSSMDNMultistreamParametricModel,
    NPSSMultistreamParametricModel,
)
from nnsvs.acoustic_models.sinsy import ResSkipF0FFConvLSTM
from nnsvs.acoustic_models.tacotron import (
    BiLSTMMDNNonAttentiveDecoder,
    BiLSTMNonAttentiveDecoder,
    MDNNonAttentiveDecoder,
    NonAttentiveDecoder,
)
from nnsvs.acoustic_models.tacotron_f0 import (
    BiLSTMResF0NonAttentiveDecoder,
    MDNResF0NonAttentiveDecoder,
    ResF0NonAttentiveDecoder,
)
from nnsvs.acoustic_models.util import predict_lf0_with_residual
from nnsvs.base import BaseModel, PredictionType
from nnsvs.layers.conv import ResnetBlock, WNConv1d
from nnsvs.mdn import MDNLayer, mdn_get_most_probable_sigma_and_mu
from nnsvs.model import TransformerEncoder, VariancePredictor
from nnsvs.util import init_weights
from torch import nn

__all__ = [
    # Non-autoregressive models
    "ResF0Conv1dResnet",
    "ResSkipF0FFConvLSTM",
    "ResF0VariancePredictor",
    "ResF0TransformerEncoder",
    # Autoregressive models
    "NonAttentiveDecoder",
    "MDNNonAttentiveDecoder",
    "BiLSTMNonAttentiveDecoder",
    "BiLSTMMDNNonAttentiveDecoder",
    "ResF0NonAttentiveDecoder",
    "MDNResF0NonAttentiveDecoder",
    "BiLSTMResF0NonAttentiveDecoder",
    # Multi-stream models
    "MultistreamSeparateF0ParametricModel",
    "NPSSMDNMultistreamParametricModel",
    "NPSSMultistreamParametricModel",
    "MultistreamSeparateF0MelModel",
    "MDNMultistreamSeparateF0MelModel",
]


class ResF0Conv1dResnet(BaseModel):
    """Conv1d + Resnet + Residual F0 prediction

    Residual F0 prediction is inspired by  :cite:t:`hono2021sinsy`.

    Args:
        in_dim (int): input dimension
        hidden_dim (int): hidden dimension
        out_dim (int): output dimension
        num_layers (int): number of layers
        in_lf0_idx (int): index of lf0 in input features
        in_lf0_min (float): minimum value of lf0 in the training data of input features
        in_lf0_max (float): maximum value of lf0 in the training data of input features
        out_lf0_idx (int): index of lf0 in output features. Typically 180.
        out_lf0_mean (float): mean of lf0 in the training data of output features
        out_lf0_scale (float): scale of lf0 in the training data of output features
        init_type (str): initialization type
        use_mdn (bool): whether to use MDN or not
        num_gaussians (int): number of gaussians in MDN
        dim_wise (bool): whether to use dimension-wise MDN or not
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=4,
        # NOTE: you must carefully set the following parameters
        in_lf0_idx=300,
        in_lf0_min=5.3936276,
        in_lf0_max=6.491111,
        out_lf0_idx=180,
        out_lf0_mean=5.953093881972361,
        out_lf0_scale=0.23435173188961034,
        init_type="none",
        use_mdn=False,
        num_gaussians=8,
        dim_wise=False,
    ):
        super().__init__()
        self.in_lf0_idx = in_lf0_idx
        self.in_lf0_min = in_lf0_min
        self.in_lf0_max = in_lf0_max
        self.out_lf0_idx = out_lf0_idx
        self.out_lf0_mean = out_lf0_mean
        self.out_lf0_scale = out_lf0_scale
        self.use_mdn = use_mdn

        model = [
            nn.ReflectionPad1d(3),
            WNConv1d(in_dim, hidden_dim, kernel_size=7, padding=0),
        ]
        for n in range(num_layers):
            model.append(ResnetBlock(hidden_dim, dilation=2 ** n))

        last_conv_out_dim = hidden_dim if use_mdn else out_dim
        model += [
            nn.LeakyReLU(0.2),
            nn.ReflectionPad1d(3),
            WNConv1d(hidden_dim, last_conv_out_dim, kernel_size=7, padding=0),
        ]
        self.model = nn.Sequential(*model)

        if self.use_mdn:
            self.mdn_layer = MDNLayer(
                in_dim=hidden_dim,
                out_dim=out_dim,
                num_gaussians=num_gaussians,
                dim_wise=dim_wise,
            )
        else:
            self.mdn_layer = None

        init_weights(self, init_type)

    def prediction_type(self):
        return (
            PredictionType.PROBABILISTIC
            if self.use_mdn
            else PredictionType.DETERMINISTIC
        )

    def has_residual_lf0_prediction(self):
        return True

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): input features
            lengths (torch.Tensor): lengths of input features
            y (torch.Tensor): output features

        Returns:
            tuple: (output features, residual log-F0)
        """
        out = self.model(x.transpose(1, 2)).transpose(1, 2)

        if self.use_mdn:
            log_pi, log_sigma, mu = self.mdn_layer(out)
        else:
            mu = out

        lf0_pred, lf0_residual = predict_lf0_with_residual(
            x,
            mu,
            self.in_lf0_idx,
            self.in_lf0_min,
            self.in_lf0_max,
            self.out_lf0_idx,
            self.out_lf0_mean,
            self.out_lf0_scale,
        )

        # Inject the predicted lf0 into the output features
        if self.use_mdn:
            mu[:, :, :, self.out_lf0_idx] = lf0_pred
        else:
            mu[:, :, self.out_lf0_idx] = lf0_pred.squeeze(-1)

        if self.use_mdn:
            return (log_pi, log_sigma, mu), lf0_residual
        else:
            return mu, lf0_residual

    def inference(self, x, lengths=None):
        """Inference step

        Args:
            x (torch.Tensor): input features
            lengths (torch.Tensor): lengths of input features

        Returns:
            tuple: (mu, sigma) if use_mdn, (output, ) otherwise
        """
        if self.use_mdn:
            (log_pi, log_sigma, mu), _ = self(x, lengths)
            sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
            return mu, sigma
        else:
            return self(x, lengths)[0]


# Will be removed in v0.1.0
ResF0Conv1dResnetMDN = partial(ResF0Conv1dResnet, use_mdn=True)


class ResF0VariancePredictor(VariancePredictor):
    """Variance predictor in :cite:t:`ren2020fastspeech` with residual F0 prediction

    Args:
        in_dim (int): the input dimension
        out_dim (int): the output dimension
        num_layers (int): the number of layers
        hidden_dim (int): the hidden dimension
        kernel_size (int): the kernel size
        dropout (float): the dropout rate
        in_lf0_idx (int): the index of the input LF0
        in_lf0_min (float): the minimum value of the input LF0
        in_lf0_max (float): the maximum value of the input LF0
        out_lf0_idx (int): the index of the output LF0
        out_lf0_mean (float): the mean value of the output LF0
        out_lf0_scale (float): the scale value of the output LF0
        init_type (str): the initialization type
        use_mdn (bool): whether to use MDN or not
        num_gaussians (int): the number of gaussians
        dim_wise (bool): whether to use dim-wise or not
    """

    def __init__(
        self,
        in_dim,
        out_dim,
        num_layers=5,
        hidden_dim=256,
        kernel_size=5,
        dropout=0.5,
        init_type="none",
        use_mdn=False,
        num_gaussians=1,
        dim_wise=False,
        # NOTE: you must carefully set the following parameters
        in_lf0_idx=300,
        in_lf0_min=5.3936276,
        in_lf0_max=6.491111,
        out_lf0_idx=180,
        out_lf0_mean=5.953093881972361,
        out_lf0_scale=0.23435173188961034,
    ):
        super().__init__(
            in_dim=in_dim,
            out_dim=out_dim,
            num_layers=num_layers,
            hidden_dim=hidden_dim,
            kernel_size=kernel_size,
            dropout=dropout,
            init_type=init_type,
            use_mdn=use_mdn,
            num_gaussians=num_gaussians,
            dim_wise=dim_wise,
        )
        self.in_lf0_idx = in_lf0_idx
        self.in_lf0_min = in_lf0_min
        self.in_lf0_max = in_lf0_max
        self.out_lf0_idx = out_lf0_idx
        self.out_lf0_mean = out_lf0_mean
        self.out_lf0_scale = out_lf0_scale

    def has_residual_lf0_prediction(self):
        return True

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): input features
            lengths (torch.Tensor): lengths of input features
            y (torch.Tensor): output features

        Returns:
            tuple: (output features, residual log-F0)
        """
        out = super().forward(x, lengths, y)
        if self.use_mdn:
            log_pi, log_sigma, mu = out
        else:
            mu = out

        lf0_pred, lf0_residual = predict_lf0_with_residual(
            x,
            mu,
            self.in_lf0_idx,
            self.in_lf0_min,
            self.in_lf0_max,
            self.out_lf0_idx,
            self.out_lf0_mean,
            self.out_lf0_scale,
        )

        # Inject the predicted lf0 into the output features
        if self.use_mdn:
            mu[:, :, :, self.out_lf0_idx] = lf0_pred
        else:
            mu[:, :, self.out_lf0_idx] = lf0_pred.squeeze(-1)

        if self.use_mdn:
            return (log_pi, log_sigma, mu), lf0_residual
        else:
            return mu, lf0_residual

    def inference(self, x, lengths=None):
        """Inference step

        Args:
            x (torch.Tensor): input features
            lengths (torch.Tensor): lengths of input features

        Returns:
            tuple: (mu, sigma) if use_mdn, (output, ) otherwise
        """
        if self.use_mdn:
            (log_pi, log_sigma, mu), _ = self(x, lengths)
            sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
            return mu, sigma
        else:
            return self(x, lengths)[0]


class ResF0TransformerEncoder(BaseModel):
    """Transformer encoder with residual f0 prediction"""

    def __init__(
        self,
        in_dim,
        out_dim,
        hidden_dim,
        attention_dim,
        num_heads=2,
        num_layers=2,
        kernel_size=3,
        dropout=0.1,
        reduction_factor=1,
        init_type="none",
        downsample_by_conv=False,
        # NOTE: you must carefully set the following parameters
        in_lf0_idx=300,
        in_lf0_min=5.3936276,
        in_lf0_max=6.491111,
        out_lf0_idx=180,
        out_lf0_mean=5.953093881972361,
        out_lf0_scale=0.23435173188961034,
    ):
        super().__init__()
        self.in_lf0_idx = in_lf0_idx
        self.in_lf0_min = in_lf0_min
        self.in_lf0_max = in_lf0_max
        self.out_lf0_idx = out_lf0_idx
        self.out_lf0_mean = out_lf0_mean
        self.out_lf0_scale = out_lf0_scale
        self.reduction_factor = reduction_factor
        self.encoder = TransformerEncoder(
            in_dim=in_dim,
            out_dim=out_dim,
            hidden_dim=hidden_dim,
            attention_dim=attention_dim,
            num_heads=num_heads,
            num_layers=num_layers,
            kernel_size=kernel_size,
            dropout=dropout,
            reduction_factor=reduction_factor,
            init_type=init_type,
            downsample_by_conv=downsample_by_conv,
        )

    def has_residual_lf0_prediction(self):
        return True

    def forward(self, x, lengths=None, y=None):
        """Forward pass

        Args:
            x (torch.Tensor): input tensor
            lengths (torch.Tensor): input sequence lengths
            y (torch.Tensor): target tensor (optional)

        Returns:
            torch.Tensor: output tensor
        """
        outs = self.encoder(x, lengths)

        lf0_pred, lf0_residual = predict_lf0_with_residual(
            x,
            outs,
            self.in_lf0_idx,
            self.in_lf0_min,
            self.in_lf0_max,
            self.out_lf0_idx,
            self.out_lf0_mean,
            self.out_lf0_scale,
        )
        outs[:, :, self.out_lf0_idx] = lf0_pred.squeeze(-1)

        return outs, lf0_residual

    def inference(self, x, lengths):
        return self(x, lengths)[0]


def LSTMEncoder(*args, **kwargs):
    from nnsvs.model import LSTMEncoder

    return LSTMEncoder(*args, **kwargs)



# The code was adapted from ttslearn https://github.com/r9y9/ttslearn
# NonAttentiveDecoder is added to the original code.
# Acknowledgement: some of the code was adapted from ESPnet
#  Copyright 2019 Nagoya University (Tomoki Hayashi)
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)

import torch
import torch.nn.functional as F
from nnsvs.base import BaseModel, PredictionType
from nnsvs.mdn import MDNLayer, mdn_get_most_probable_sigma_and_mu, mdn_get_sample
from nnsvs.util import init_weights
from torch import nn


def decoder_init(m):
    if isinstance(m, nn.Conv1d):
        nn.init.xavier_uniform_(m.weight, nn.init.calculate_gain("tanh"))


class ZoneOutCell(nn.Module):
    def __init__(self, cell, zoneout=0.1):
        super().__init__()
        self.cell = cell
        self.hidden_size = cell.hidden_size
        self.zoneout = zoneout

    def forward(self, inputs, hidden):
        next_hidden = self.cell(inputs, hidden)
        next_hidden = self._zoneout(hidden, next_hidden, self.zoneout)
        return next_hidden

    def _zoneout(self, h, next_h, prob):
        h_0, c_0 = h
        h_1, c_1 = next_h
        h_1 = self._apply_zoneout(h_0, h_1, prob)
        c_1 = self._apply_zoneout(c_0, c_1, prob)
        return h_1, c_1

    def _apply_zoneout(self, h, next_h, prob):
        if self.training:
            if prob > 0.0:
                mask = h.new(*h.size()).bernoulli_(prob)
            else:
                mask = 0
            return mask * h + (1 - mask) * next_h
        else:
            return prob * h + (1 - prob) * next_h


class Prenet(nn.Module):
    """Pre-Net of Tacotron.

    Args:
        in_dim (int) : dimension of input
        layers (int) : number of pre-net layers
        hidden_dim (int) : dimension of hidden layer
        dropout (float) : dropout rate
    """

    def __init__(
        self, in_dim, layers=2, hidden_dim=256, dropout=0.5, eval_dropout=True
    ):
        super().__init__()
        self.dropout = dropout
        self.eval_dropout = eval_dropout
        prenet = nn.ModuleList()
        for layer in range(layers):
            prenet += [
                nn.Linear(in_dim if layer == 0 else hidden_dim, hidden_dim),
                nn.ReLU(),
            ]
        self.prenet = nn.Sequential(*prenet)

    def forward(self, x):
        """Forward step

        Args:
            x (torch.Tensor) : input tensor

        Returns:
            torch.Tensor : output tensor
        """
        for layer in self.prenet:
            if self.eval_dropout:
                x = F.dropout(layer(x), self.dropout, training=True)
            else:
                x = F.dropout(layer(x), self.dropout, training=self.training)
        return x


class NonAttentiveDecoder(BaseModel):
    """Decoder of Tacotron w/o attention mechanism

    Args:
        in_dim (int) : dimension of encoder hidden layer
        out_dim (int) : dimension of output
        layers (int) : number of LSTM layers
        hidden_dim (int) : dimension of hidden layer
        prenet_layers (int) : number of pre-net layers
        prenet_hidden_dim (int) : dimension of pre-net hidden layer
        prenet_dropout (float) : dropout rate of pre-net
        zoneout (float) : zoneout rate
        reduction_factor (int) : reduction factor
        attention_hidden_dim (int) : dimension of attention hidden layer
        attention_conv_channel (int) : number of attention convolution channels
        attention_conv_kernel_size (int) : kernel size of attention convolution
        downsample_by_conv (bool) : if True, downsample by convolution
        initial_value (float) : initial value for the autoregressive decoder.
    """

    def __init__(
        self,
        in_dim=512,
        out_dim=80,
        layers=2,
        hidden_dim=1024,
        prenet_layers=2,
        prenet_hidden_dim=256,
        prenet_dropout=0.5,
        zoneout=0.1,
        reduction_factor=1,
        downsample_by_conv=False,
        init_type="none",
        eval_dropout=True,
        prenet_noise_std=0.0,
        initial_value=0.0,
    ):
        super().__init__()
        self.out_dim = out_dim
        self.reduction_factor = reduction_factor
        self.prenet_dropout = prenet_dropout
        self.prenet_noise_std = prenet_noise_std
        self.initial_value = initial_value

        if prenet_layers > 0:
            self.prenet = Prenet(
                out_dim,
                prenet_layers,
                prenet_hidden_dim,
                prenet_dropout,
                eval_dropout=eval_dropout,
            )
            lstm_in_dim = in_dim + prenet_hidden_dim
        else:
            self.prenet = None
            prenet_hidden_dim = 0
            lstm_in_dim = in_dim + out_dim

        self.lstm = nn.ModuleList()
        for layer in range(layers):
            lstm = nn.LSTMCell(
                lstm_in_dim if layer == 0 else hidden_dim,
                hidden_dim,
            )
            self.lstm += [ZoneOutCell(lstm, zoneout)]

        proj_in_dim = in_dim + hidden_dim
        self.feat_out = nn.Linear(proj_in_dim, out_dim * reduction_factor, bias=False)

        if reduction_factor > 1 and downsample_by_conv:
            self.conv_downsample = nn.Conv1d(
                in_dim,
                in_dim,
                kernel_size=reduction_factor,
                stride=reduction_factor,
                groups=in_dim,
            )
        else:
            self.conv_downsample = None

        init_weights(self, init_type)

    def _zero_state(self, hs):
        init_hs = hs.new_zeros(hs.size(0), self.lstm[0].hidden_size)
        return init_hs

    def is_autoregressive(self):
        return True

    def forward(self, encoder_outs, in_lens, decoder_targets=None):
        """Forward step

        Args:
            encoder_outs (torch.Tensor): encoder outputs (B, T, C)
            in_lens (torch.Tensor): input lengths
            decoder_targets (torch.Tensor): decoder targets for teacher-forcing. (B, T, C)

        Returns:
            torch.Tensor: the output (B, C, T)
        """
        is_inference = decoder_targets is None
        if not is_inference:
            assert encoder_outs.shape[1] == decoder_targets.shape[1]

        # Adjust number of frames according to the reduction factor
        # (B, Lmax, out_dim) ->  (B, Lmax/r, out_dim)
        if self.reduction_factor > 1 and not is_inference:
            decoder_targets = decoder_targets[
                :, self.reduction_factor - 1 :: self.reduction_factor
            ]
        if self.reduction_factor > 1:
            if self.conv_downsample is not None:
                encoder_outs = self.conv_downsample(
                    encoder_outs.transpose(1, 2)
                ).transpose(1, 2)
            else:
                encoder_outs = encoder_outs[
                    :, self.reduction_factor - 1 :: self.reduction_factor
                ]

        h_list, c_list = [], []
        for _ in range(len(self.lstm)):
            h_list.append(self._zero_state(encoder_outs))
            c_list.append(self._zero_state(encoder_outs))

        go_frame = (
            encoder_outs.new_zeros(encoder_outs.size(0), self.out_dim)
            + self.initial_value
        )
        prev_out = go_frame

        if not is_inference and self.prenet is not None:
            prenet_outs = self.prenet(decoder_targets)

        outs = []
        for t in range(encoder_outs.shape[1]):
            # Pre-Net
            if self.prenet is not None:
                if is_inference:
                    prenet_out = self.prenet(prev_out)
                else:
                    prenet_out = prenet_outs[:, t, :]
            elif self.prenet_noise_std > 0:
                prenet_out = (
                    prev_out + torch.randn_like(prev_out) * self.prenet_noise_std
                )
            else:
                prenet_out = F.dropout(prev_out, self.prenet_dropout, training=True)

            # LSTM
            xs = torch.cat([encoder_outs[:, t], prenet_out], dim=1)
            h_list[0], c_list[0] = self.lstm[0](xs, (h_list[0], c_list[0]))
            for i in range(1, len(self.lstm)):
                h_list[i], c_list[i] = self.lstm[i](
                    h_list[i - 1], (h_list[i], c_list[i])
                )
            # Output
            hcs = torch.cat([h_list[-1], encoder_outs[:, t]], dim=1)
            outs.append(self.feat_out(hcs).view(encoder_outs.size(0), self.out_dim, -1))

            # Update decoder input for the next time step
            if is_inference:
                prev_out = outs[-1][:, :, -1]  # (1, out_dim)
            else:
                # Teacher forcing
                prev_out = decoder_targets[:, t, :]

        outs = torch.cat(outs, dim=2)  # (B, out_dim, Lmax)

        if self.reduction_factor > 1:
            outs = outs.view(outs.size(0), self.out_dim, -1)  # (B, out_dim, Lmax)

        # (B, C, T) -> (B, T, C)
        return outs.transpose(1, 2)


class MDNNonAttentiveDecoder(BaseModel):
    """Non-atteive decoder with MDN

    Each decoder step outputs the parameters of MDN.

    Args:
        in_dim (int): input dimension
        out_dim (int): output dimension
        layers (int): number of LSTM layers
        hidden_dim (int): hidden dimension
        prenet_layers (int): number of prenet layers
        prenet_hidden_dim (int): prenet hidden dimension
        prenet_dropout (float): prenet dropout rate
        zoneout (float): zoneout rate
        reduction_factor (int): reduction factor
        downsample_by_conv (bool): if True, use conv1d to downsample the input
        num_gaussians (int): number of Gaussians
        sampling_mode (str): sampling mode
        init_type (str): initialization type
        eval_dropout (bool): if True, use dropout in evaluation
        initial_value (float) : initial value for the autoregressive decoder.
    """

    def __init__(
        self,
        in_dim=512,
        out_dim=80,
        layers=2,
        hidden_dim=1024,
        prenet_layers=2,
        prenet_hidden_dim=256,
        prenet_dropout=0.5,
        zoneout=0.1,
        reduction_factor=1,
        downsample_by_conv=False,
        num_gaussians=8,
        sampling_mode="mean",
        init_type="none",
        eval_dropout=True,
        prenet_noise_std=0.0,
        initial_value=0.0,
    ):
        super().__init__()
        self.out_dim = out_dim
        self.reduction_factor = reduction_factor
        self.prenet_dropout = prenet_dropout
        self.prenet_noise_std = prenet_noise_std
        self.num_gaussians = num_gaussians
        self.sampling_mode = sampling_mode
        assert sampling_mode in ["mean", "random"]
        self.initial_value = initial_value

        if prenet_layers > 0:
            self.prenet = Prenet(
                out_dim,
                prenet_layers,
                prenet_hidden_dim,
                prenet_dropout,
                eval_dropout=eval_dropout,
            )
            lstm_in_dim = in_dim + prenet_hidden_dim
        else:
            self.prenet = None
            prenet_hidden_dim = 0
            lstm_in_dim = in_dim + out_dim

        self.lstm = nn.ModuleList()
        for layer in range(layers):
            lstm = nn.LSTMCell(
                lstm_in_dim if layer == 0 else hidden_dim,
                hidden_dim,
            )
            self.lstm += [ZoneOutCell(lstm, zoneout)]

        proj_in_dim = in_dim + hidden_dim
        self.feat_out = MDNLayer(
            proj_in_dim,
            out_dim * reduction_factor,
            num_gaussians=num_gaussians,
            dim_wise=True,
        )

        if reduction_factor > 1 and downsample_by_conv:
            self.conv_downsample = nn.Conv1d(
                in_dim,
                in_dim,
                kernel_size=reduction_factor,
                stride=reduction_factor,
                groups=in_dim,
            )
        else:
            self.conv_downsample = None

        init_weights(self, init_type)

    def _zero_state(self, hs):
        init_hs = hs.new_zeros(hs.size(0), self.lstm[0].hidden_size)
        return init_hs

    def is_autoregressive(self):
        return True

    def prediction_type(self):
        return PredictionType.PROBABILISTIC

    def forward(self, encoder_outs, in_lens, decoder_targets=None):
        is_inference = decoder_targets is None
        if not is_inference:
            assert encoder_outs.shape[1] == decoder_targets.shape[1]

        # Adjust number of frames according to the reduction factor
        # (B, Lmax, out_dim) ->  (B, Lmax/r, out_dim)
        if self.reduction_factor > 1 and not is_inference:
            decoder_targets = decoder_targets[
                :, self.reduction_factor - 1 :: self.reduction_factor
            ]
        if self.reduction_factor > 1:
            if self.conv_downsample is not None:
                encoder_outs = self.conv_downsample(
                    encoder_outs.transpose(1, 2)
                ).transpose(1, 2)
            else:
                encoder_outs = encoder_outs[
                    :, self.reduction_factor - 1 :: self.reduction_factor
                ]

        h_list, c_list = [], []
        for _ in range(len(self.lstm)):
            h_list.append(self._zero_state(encoder_outs))
            c_list.append(self._zero_state(encoder_outs))

        go_frame = (
            encoder_outs.new_zeros(encoder_outs.size(0), self.out_dim)
            + self.initial_value
        )
        prev_out = go_frame

        if not is_inference and self.prenet is not None:
            prenet_outs = self.prenet(decoder_targets)

        mus = []
        log_pis = []
        log_sigmas = []
        # NOTE: only used for inference
        mus_inf = []

        for t in range(encoder_outs.shape[1]):
            # Pre-Net
            if self.prenet is not None:
                if is_inference:
                    prenet_out = self.prenet(prev_out)
                else:
                    prenet_out = prenet_outs[:, t, :]
            elif self.prenet_noise_std > 0:
                prenet_out = (
                    prev_out + torch.randn_like(prev_out) * self.prenet_noise_std
                )
            else:
                prenet_out = F.dropout(prev_out, self.prenet_dropout, training=True)

            # LSTM
            xs = torch.cat([encoder_outs[:, t], prenet_out], dim=1)
            h_list[0], c_list[0] = self.lstm[0](xs, (h_list[0], c_list[0]))
            for i in range(1, len(self.lstm)):
                h_list[i], c_list[i] = self.lstm[i](
                    h_list[i - 1], (h_list[i], c_list[i])
                )
            # Output
            hcs = torch.cat([h_list[-1], encoder_outs[:, t]], dim=1)
            log_pi, log_sigma, mu = self.feat_out(hcs.unsqueeze(1))

            # (B, 1, num_gaussians, out_dim*reduction_factor)
            # -> (B, reduction_factor, num_gaussians, out_dim)
            log_pi = (
                log_pi.transpose(1, 2)
                .view(encoder_outs.size(0), self.num_gaussians, -1, self.out_dim)
                .transpose(1, 2)
            )
            log_sigma = (
                log_sigma.transpose(1, 2)
                .view(encoder_outs.size(0), self.num_gaussians, -1, self.out_dim)
                .transpose(1, 2)
            )
            mu = (
                mu.transpose(1, 2)
                .view(encoder_outs.size(0), self.num_gaussians, -1, self.out_dim)
                .transpose(1, 2)
            )

            mus.append(mu)
            log_pis.append(log_pi)
            log_sigmas.append(log_sigma)

            # Update decoder input for the next time step
            if is_inference:
                # (B, reduction_factor, out_dim)
                if self.sampling_mode == "mean":
                    _, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
                elif self.sampling_mode == "random":
                    mu = mdn_get_sample(log_pi, log_sigma, mu)

                # Feed last sample for the feedback loop
                prev_out = mu[:, -1]

                mus_inf.append(mu)
            else:
                # Teacher forcing
                prev_out = decoder_targets[:, t, :]

        mus = torch.cat(mus, dim=1)  # (B, out_dim, Lmax)
        log_pis = torch.cat(log_pis, dim=1)  # (B, out_dim, Lmax)
        log_sigmas = torch.cat(log_sigmas, dim=1)  # (B, out_dim, Lmax)

        if is_inference:
            mu = torch.cat(mus_inf, dim=1)
            # TODO: may need to track sigma. For now we only use mu
            return mu, mu
        else:
            return log_pis, log_sigmas, mus
